"""Functions for calculating evapotranspiration."""

import numpy as np
import numpy.typing as npt
from numba import njit

from geb.geb_types import ArrayFloat32

from ..landcovers import (
    FOREST,
    GRASSLAND_LIKE,
    OPEN_WATER,
    PADDY_IRRIGATED,
    SEALED,
)

# TODO: Load this dynamically as global var (see model.py)
N_SOIL_LAYERS: int = 6


@njit(cache=True, inline="always")
def get_critical_soil_moisture_content(
    p: np.float32,
    wfc_m: ArrayFloat32,
    wwp_m: ArrayFloat32,
) -> ArrayFloat32:
    """Calculate the critical soil moisture content.

    The critical soil moisture content is defined as the quantity of stored soil moisture below
    which water uptake is impaired and the crop begins to close its stomata. It is not a fixed
    value. Restriction of water uptake due to water stress starts at a higher water content
    when the potential transpiration rate is higher" (Van Diepen et al., 1988: WOFOST 6.0, p.86).

    A higher p value means that the critical soil moisture content is higher, i.e. the plant can
    extract water from the soil at a lower soil moisture content. Thus when p is 1 the critical
    soil moisture content is equal to the wilting point, and when p is 0 the critical soil moisture
    content is equal to the field capacity.

    Args:
        p: The fraction of easily available soil water, between 0 and 1.
        wfc_m: The field capacity in m.
        wwp_m: The wilting point in m.

    Returns:
        The critical soil moisture content in m.

    """
    return (np.float32(1) - p) * (wfc_m - wwp_m) + wwp_m


@njit(cache=True, inline="always")
def get_root_mass_ratios(
    root_depth_m: np.float32,
    root_ratios: ArrayFloat32,
    soil_layer_height_m: ArrayFloat32,
) -> ArrayFloat32:
    """Calculate the root mass ratios for each soil layer assuming a triangular root distribution.

    Args:
        root_depth_m: The effective root depth in meters.
        root_ratios: The root ratios for each soil layer, where each ratio
        soil_layer_height_m: The height of each soil layer in meters.

    Returns:
        A numpy array of root mass ratios for each soil layer. The total root mass ratio is 1.
    """
    current_depth = np.float32(0)
    root_mass_ratio = np.zeros_like(soil_layer_height_m, dtype=np.float32)

    root_triangle_area = np.float32(0.5) * root_depth_m * root_depth_m

    for layer in range(N_SOIL_LAYERS):
        if root_ratios[layer] == np.float32(0):
            break  # No roots in this layer (so also not in remaining), skip further calculations

        triangle_block_height = soil_layer_height_m[layer] * root_ratios[layer]
        triangle_block_center_point_depth = triangle_block_height / 2 + current_depth
        triangle_block_width = root_depth_m - triangle_block_center_point_depth
        root_mass_ratio[layer] = (
            triangle_block_height * triangle_block_width / root_triangle_area
        )

        current_depth += soil_layer_height_m[layer]

    return root_mass_ratio


@njit(cache=True, inline="always")
def get_transpiration_factor(
    w_m: np.float32, wwp_m: np.float32, wcrit_m: np.float32
) -> np.float32:
    """Calculate the transpiration factor based on the available water in the soil.

    A factor of 1 means that there is no water stress and the plant can transpire at its potential rate.
    A factor of 0 means that the plant is at wilting point and cannot transpire.

    Args:
        w_m: The soil water content in m.
        wwp_m: The wilting point in m.
        wcrit_m: The critical soil moisture content in m.

    Returns:
        The transpiration factor between 0 and 1.
    """
    nominator = w_m - wwp_m
    denominator = wcrit_m - wwp_m
    if denominator == np.float32(0):
        if nominator > np.float32(0):
            return np.float32(1)
        else:
            return np.float32(0)
    factor = nominator / denominator
    if factor < np.float32(0):
        return np.float32(0)
    if factor > np.float32(1):
        return np.float32(1)
    return factor


@njit(cache=True, inline="always")
def get_root_ratios(
    root_depth_m: np.float32, soil_layer_height_m: npt.NDArray[np.float32]
) -> npt.NDArray[np.float32]:
    """Calculate the root ratios for each soil layer based on the effective root depth.

    Args:
        root_depth_m: The effective root depth in meters.
        soil_layer_height_m: The height of each soil layer in meters.

    Returns:
        A numpy array of root ratios for each soil layer, where each ratio
        is the proportion of the soil layer that has roots in it.
    """
    remaining_root_depth = root_depth_m
    root_ratios = np.zeros_like(soil_layer_height_m)
    for layer in range(N_SOIL_LAYERS):
        root_ratios[layer] = min(
            remaining_root_depth / soil_layer_height_m[layer], np.float32(1)
        )
        remaining_root_depth -= soil_layer_height_m[layer]
        if remaining_root_depth < np.float32(0):
            return root_ratios
    return root_ratios


@njit(cache=True, inline="always")
def get_fraction_easily_available_soil_water(
    crop_group_number: np.float32,
    reference_evapotranspiration_grass_full_day_m: np.float32,
) -> np.float32:
    """Calculate the fraction of easily available soil water.

    Calculation is based on crop group number and potential evapotranspiration
    following Van Diepen et al., 1988: WOFOST 6.0, Theory and Algorithms p.87.

    Args:
        crop_group_number: The crop group number is a indicator of adaptation to dry climate,
            Van Diepen et al., 1988: WOFOST 6.0, Theory and Algorithms p.87
        reference_evapotranspiration_grass_full_day_m: Potential evapotranspiration in m for a full day.

    Returns:
        The fraction of easily available soil water, p is closer to 0 if evapo is bigger and cropgroup is smaller
    """
    potential_evapotranspiration_cm: np.float32 = (
        reference_evapotranspiration_grass_full_day_m * np.float32(100)
    )

    p: np.float32 = np.float32(1) / (
        np.float32(0.76) + np.float32(1.5) * potential_evapotranspiration_cm
    ) - np.float32(0.1) * (np.float32(5) - crop_group_number)

    # Additional correction for crop groups 1 and 2
    if crop_group_number <= np.float32(2.5):
        p: np.float32 = np.float32(p) + (
            potential_evapotranspiration_cm - np.float32(0.6)
        ) / (crop_group_number * (crop_group_number + np.float32(3.0)))

    if p < np.float32(0):
        p: np.float32 = np.float32(0)
    if p > np.float32(1):
        p: np.float32 = np.float32(1)
    return p


@njit(inline="always", cache=True)
def get_transpiration_factor_per_layer(
    soil_layer_height_m: npt.NDArray[np.float32],
    effective_root_depth_m: np.float32,
    w_m: npt.NDArray[np.float32],
    wfc_m: npt.NDArray[np.float32],
    wwp_m: npt.NDArray[np.float32],
    p: np.float32,
    correct_root_mass: bool = True,
) -> npt.NDArray[np.float32]:
    """Calculate the transpiration factor for each soil layer based on the available water and root ratios.

    Args:
        soil_layer_height_m: height of each soil layer in meters.
        effective_root_depth_m: root depth in meters, which is the maximum depth where roots can extract water.
        w_m: soil water content in each layer in m.
        wfc_m: field capacity in each layer in m.
        wwp_m: wilting point in each layer in m.
        p: ratio of easily available soil water, which is a measure of how easily plants can extract water from the soil.
        correct_root_mass: If True, the root mass ratio is corrected assuming a triangular root distribution.

    Returns:
        A numpy array of transpiration factors [0-1] for each soil layer, where each factor
            is the proportion of the available water in that layer that can be used
            for transpiration. If plenty of water available, sums to 1, else less than 1.
    """
    root_ratios = get_root_ratios(
        effective_root_depth_m,
        soil_layer_height_m,
    )

    if correct_root_mass:
        root_mass_ratio_per_layer = get_root_mass_ratios(
            effective_root_depth_m,
            root_ratios,
            soil_layer_height_m,
        )
    else:
        root_mass_ratio_per_layer = (
            soil_layer_height_m * root_ratios / effective_root_depth_m
        )

    w_available = (w_m * root_ratios).sum()
    wfc_available = (wfc_m * root_ratios).sum()
    wwp_available = (wwp_m * root_ratios).sum()

    # Compute total water stress factor based on the available water in the soil
    critical_soil_moisture_content = get_critical_soil_moisture_content(
        p, wfc_available, wwp_available
    )
    transpiration_factor_total = get_transpiration_factor(
        w_available, wwp_available, critical_soil_moisture_content
    )
    if transpiration_factor_total == np.float32(0):
        return np.zeros_like(w_m)

    transpiration_factor_per_layer = np.zeros_like(w_m)

    for layer in range(N_SOIL_LAYERS):
        critical_soil_moisture_content_layer = get_critical_soil_moisture_content(
            p, wfc_m[layer], wwp_m[layer]
        )

        transpiration_factor_per_layer[layer] = (
            get_transpiration_factor(
                w_m[layer], wwp_m[layer], critical_soil_moisture_content_layer
            )
            * root_mass_ratio_per_layer[layer]
        )

    transpiration_factor_per_layer = (
        transpiration_factor_per_layer
        / transpiration_factor_per_layer.sum()
        * transpiration_factor_total
    )

    return transpiration_factor_per_layer


@njit(cache=True, inline="always")
def calculate_transpiration(
    soil_is_frozen: bool,
    wwp_m: npt.NDArray[np.float32],  # [m]
    wfc_m: npt.NDArray[np.float32],  # [m]
    wres_m: npt.NDArray[np.float32],  # [m]
    soil_layer_height_m: npt.NDArray[np.float32],  # [m]
    land_use_type: np.int32,
    root_depth_m: np.float32,  # [m]
    crop_map: np.int32,
    crop_group_forest: np.float32,
    crop_group_grassland_like: np.float32,
    potential_transpiration_m: np.float32,  # [m]
    reference_evapotranspiration_grass_m_hour: np.float32,  # [m]
    crop_group_number_per_group: npt.NDArray[np.float32],
    w_m: npt.NDArray[np.float32],  # [m]
    topwater_m: np.float32,  # [m]
    minimum_effective_root_depth_m: np.float32,  # [m]
) -> tuple[np.float32, np.float32]:
    """Calculate transpiration for a single soil cell.

    Args:
        soil_is_frozen: Boolean indicating whether the soil is frozen.
        wwp_m: Wilting point soil moisture content [m], shape (N_SOIL_LAYERS,).
        wfc_m: Field capacity soil moisture content [m], shape (N_SOIL_LAYERS,).
        wres_m: Residual soil moisture content [m], shape (N_SOIL_LAYERS,).
        soil_layer_height_m: Height of each soil layer [m], shape (N_SOIL_LAYERS,).
        land_use_type: Land use type of the hydrological response unit.
        root_depth_m: The root depth [m].
        crop_map: Crop map indicating the crop type for the hydrological response unit. -1 indicates no crop.
        crop_group_forest: Crop group numbers for forest (see WOFOST 6.0).
        crop_group_grassland_like: Crop group numbers for grassland-like areas (see WOFOST 6.0).
        potential_transpiration_m: Potential transpiration [m].
        reference_evapotranspiration_grass_m_hour: Reference evapotranspiration for grass [m/hour]. This is an hourly value that will be converted to a daily value for the calculation.
        crop_group_number_per_group: Crop group numbers for each crop type.
        w_m: Soil water content [m], shape (N_SOIL_LAYERS,).
        topwater_m: Topwater [m], which is the water available for evaporation and transpiration for paddy irrigated fields.
        minimum_effective_root_depth_m: Minimum effective root depth [m], used to ensure that the effective root depth is not less than this value. Crops can extract water up to this depth.

    Returns:
        A tuple containing:
            - transpiration: The actual transpiration [m] for the cell.
            - topwater_m: Updated topwater [m] after transpiration.
    """
    transpiration: np.float32 = np.float32(0.0)

    remaining_potential_transpiration: np.float32 = potential_transpiration_m
    if land_use_type == PADDY_IRRIGATED:
        transpiration_from_topwater: np.float32 = min(
            topwater_m, remaining_potential_transpiration
        )
        remaining_potential_transpiration -= transpiration_from_topwater
        topwater_m -= transpiration_from_topwater
        transpiration += transpiration_from_topwater

    if not soil_is_frozen:
        # get group group numbers for natural areas
        if land_use_type == FOREST:
            crop_group_number: np.float32 = crop_group_forest
        elif land_use_type == GRASSLAND_LIKE:
            crop_group_number: np.float32 = crop_group_grassland_like
        else:  #
            crop_group_number: np.float32 = crop_group_number_per_group[crop_map]

        # vegetation-specific factor for easily available soil water
        fraction_easily_available_soil_water: np.float32 = get_fraction_easily_available_soil_water(
            crop_group_number=crop_group_number,
            reference_evapotranspiration_grass_full_day_m=reference_evapotranspiration_grass_m_hour
            * np.float32(24.0),
        )

        effective_root_depth: np.float32 = np.maximum(
            minimum_effective_root_depth_m, root_depth_m
        )
        fraction_easily_available_soil_water = np.float32(
            fraction_easily_available_soil_water
        )

        transpiration_factor_per_layer: npt.NDArray[np.float32] = (
            get_transpiration_factor_per_layer(
                soil_layer_height_m,
                effective_root_depth,
                w_m,
                wfc_m,
                wwp_m,
                fraction_easily_available_soil_water,
            )
        )

        for layer in range(N_SOIL_LAYERS):
            transpiration_layer: np.float32 = (
                remaining_potential_transpiration
                * transpiration_factor_per_layer[layer]
            )
            # limit the transpiration to the available water in the soil
            transpiration_layer = min(
                transpiration_layer,
                w_m[layer] - wres_m[layer],
            )

            w_m[layer] -= transpiration_layer
            w_m[layer] = max(
                w_m[layer], wres_m[layer]
            )  # soil moisture can never be lower than wres
            transpiration += transpiration_layer

    return transpiration, topwater_m


@njit(cache=True, inline="always")
def calculate_bare_soil_evaporation(
    soil_is_frozen: bool,
    land_use_type: np.int32,
    potential_direct_evaporation_m: np.float32,
    open_water_evaporation_m: np.float32,
    w_m: npt.NDArray[np.float32],
    wres_m: npt.NDArray[np.float32],
    wfc_m: npt.NDArray[np.float32],
    unsaturated_hydraulic_conductivity_m_per_hour: np.float32,
) -> np.float32:
    """Calculate bare soil evaporation for a single soil cell.

    Args:
        soil_is_frozen: Indicate whether the soil is frozen.
        land_use_type: Land use type (-).
        potential_direct_evaporation_m: Potential direct evaporation (soil/water) (m).
        open_water_evaporation_m: Actual open water evaporation (m).
        w_m: Soil water content (m), shape (N_SOIL_LAYERS,).
        wres_m: Residual soil moisture content (m), shape (N_SOIL_LAYERS,).
        wfc_m: Field capacity soil moisture content (m), shape (N_SOIL_LAYERS,).
        unsaturated_hydraulic_conductivity_m_per_hour: Unsaturated hydraulic conductivity (m/h).

    Returns:
        Actual bare soil evaporation (m).
    """
    # Bare soil evaporation is only applied for natural (non-watery, non-paddy, non-sealed) areas.
    if (
        not soil_is_frozen
        and land_use_type != PADDY_IRRIGATED
        and land_use_type != OPEN_WATER
        and land_use_type != SEALED
    ):
        # Apply an additional correction due to dust mulch
        # reduction based on relative moisture of the top layer.
        relative_moisture = max(
            np.float32(0.0),
            (w_m[0] - wres_m[0]) / (wfc_m[0] - wres_m[0]),
        )
        dust_mulch_reduction = min(np.float32(1.0), relative_moisture**2)

        potential_direct_evaporation_m = (
            potential_direct_evaporation_m * dust_mulch_reduction
        )

        # Limit potential evaporation by the unsaturated hydraulic conductivity
        # This accounts for the reduced ability of the soil to transport water to the surface
        potential_direct_evaporation_m = min(
            potential_direct_evaporation_m,
            unsaturated_hydraulic_conductivity_m_per_hour,
        )

        # Subtract open water evaporation (though it's expected to be 0 for these land uses)
        actual_bare_soil_evaporation = min(
            max(
                np.float32(0),
                potential_direct_evaporation_m - open_water_evaporation_m,
            ),
            max(w_m[0] - wres_m[0], np.float32(0)),
        )
        # Update soil moisture
        w_m[0] -= actual_bare_soil_evaporation
        w_m[0] = max(w_m[0], wres_m[0])
    else:
        actual_bare_soil_evaporation = np.float32(0)

    return actual_bare_soil_evaporation



"""WhyMap adapter."""

from __future__ import annotations

import shutil
import zipfile
from pathlib import Path
from typing import Any

import geopandas as gpd

from geb.workflows.io import fetch_and_save

from .base import Adapter


class WhyMap(Adapter):
    """Adapter for the WhyMap dataset.

    This datasets provides global information on aquifer properties.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the WhyMap adapter.

        Args:
            *args: Arguments to pass to the Adapter constructor.
            **kwargs: Keyword arguments to pass to the Adapter constructor.

        """
        super().__init__(*args, **kwargs)

    def fetch(self, url: str, *args: Any, **kwargs: Any) -> WhyMap:
        """Fetch the WhyMap dataset.

        Args:
            url: URL to fetch the dataset from.
            *args: Additional arguments to pass to fetch_and_save.
            **kwargs: Additional keyword arguments to pass to fetch_and_save.

        Returns:
            The WhyMap adapter.
        """
        if not self.is_ready:
            download_path: Path = self.root / url.split("/")[-1]
            fetch_and_save(url, download_path)

            with zipfile.ZipFile(download_path, "r") as zip_ref:
                zip_ref.extractall(self.root)

            gdf: gpd.GeoDataFrame = gpd.read_file(
                filename=self.root
                / "WHYMAP_GWR"
                / "shp"
                / "whymap_GW_aquifers_v1_poly.shp"
            )
            gdf.to_parquet(self.path, index=False)

            shutil.rmtree(path=self.root / "WHYMAP_GWR")

            download_path.unlink()

        return self



"""This module contains classes and functions to handle Hydrological Response Units (HRUs) and grid cells."""

from __future__ import annotations

import math
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, cast, overload

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
import xarray as xr
import zarr
from affine import Affine
from numba import njit
from numpy.typing import DTypeLike
from scipy.spatial import KDTree

from geb.geb_types import (
    AnyDArrayWithScalar,
    Array,
    ArrayBool,
    ArrayFloat32,
    ArrayFloat64,
    ArrayInt32,
    ArrayInt64,
    ArrayWithScalar,
    T_ArrayNumber,
    T_OneorTwoDArray,
    ThreeDArray,
    ThreeDArrayFloat32,
    ThreeDArrayWithScalar,
    TwoDArray,
    TwoDArrayBool,
    TwoDArrayFloat32,
    TwoDArrayFloat64,
    TwoDArrayInt32,
    TwoDArrayWithScalar,
)
from geb.store import Bucket
from geb.workflows.io import read_grid, read_zarr
from geb.workflows.raster import compress, decompress_with_mask

if TYPE_CHECKING:
    from geb.model import GEBModel


def determine_nearest_river_cell(
    upstream_area: TwoDArrayFloat32,
    HRU_to_grid: ArrayInt32,
    mask: TwoDArrayBool,
    threshold_m2: float | int,
) -> ArrayInt32:
    """This function finds the nearest river cell to each HRU.

    It does so by first selecting the rivers, by checking if the upstream area is
    above a certain threshold. then for each grid cell, it finds the nearest
    river cell. Finally, it maps the nearest river cell to each HRU.

    Args:
        upstream_area: 2D-array of upstream area in m².
        HRU_to_grid: Array mapping HRUs to grid cells.
        mask: Mask of the study area.
        threshold_m2: Threshold in m² to consider a cell as a river.

    Returns:
        For each HRU, the index of the nearest river cell in the valid grid cells.
    """
    valid_indices = np.argwhere(~mask)
    valid_values = upstream_area[~mask]

    grid_cells_above_threshold_mask: npt.NDArray[np.bool_] = valid_values > threshold_m2
    grid_cells_above_threshold_indices: npt.NDArray[np.int64] = valid_indices[
        grid_cells_above_threshold_mask
    ]
    grid_cells_above_threshold_indices_in_valid: npt.NDArray[np.int64] = np.flatnonzero(
        grid_cells_above_threshold_mask
    )

    tree: KDTree = KDTree(grid_cells_above_threshold_indices)
    distances, indices_in_above = tree.query(valid_indices)

    nearest_indices_in_valid: ArrayInt32 = (
        grid_cells_above_threshold_indices_in_valid[indices_in_above]
    ).astype(np.int32)

    assert nearest_indices_in_valid.max() < (~mask).sum()

    return nearest_indices_in_valid[HRU_to_grid]


def load_water_demand_xr(filepath: str | Path) -> xr.Dataset:
    """Load a water demand dataset from disk.

    Args:
        filepath: Path to the water demand dataset file.

    Returns:
        An xarray Dataset containing the water demand data.
    """
    return xr.open_dataset(
        filepath,
        engine="zarr",
        consolidated=False,
    )


@njit(cache=True)
def to_grid(
    data: T_ArrayNumber,
    grid_to_HRU: ArrayInt32,
    land_use_ratio: ArrayFloat32,
    fn: str = "weightedmean",
) -> T_ArrayNumber:
    """Numba helper function to convert from HRU to grid.

    Args:
        data: The HRU data to be converted (1D array).
        grid_to_HRU: Array of size of the compressed grid cells. Each value maps to the index of the first unit of the next cell.
        land_use_ratio: Relative size of HRU to grid.
        fn: Name of function to apply to data. In most cases, several HRUs are combined into one grid unit, so a function must be applied. Choose from `mean`, `sum`, `nansum`, `max` and `min`.

    Returns:
        ouput_data: Data converted to grid.
    """
    output_data = np.empty(grid_to_HRU.size, dtype=data.dtype)

    assert grid_to_HRU[0] != 0, (
        "First value of grid_to_HRU cannot be 0. This would mean that the first HRU is empty."
    )
    assert grid_to_HRU[-1] == land_use_ratio.size, (
        "The last value of grid_to_HRU must be equal to the size of land_use_ratio. Otherwise, the last HRU would not be used."
    )

    prev_index = 0
    for i in range(grid_to_HRU.size):
        cell_index = grid_to_HRU[i]
        if fn == "weightedmean":
            values = data[prev_index:cell_index]
            weights = land_use_ratio[prev_index:cell_index]
            output_data[i] = (values * weights).sum() / weights.sum()
        elif fn == "weightednanmean":
            values = data[prev_index:cell_index]
            weights = land_use_ratio[prev_index:cell_index]
            weights = weights[~np.isnan(values)]
            values = values[~np.isnan(values)]
            if values.size == 0:
                output_data[i] = np.nan
            else:
                output_data[i] = (values * weights).sum() / weights.sum()
        elif fn == "sum":
            output_data[i] = np.sum(data[prev_index:cell_index])
        elif fn == "nansum":
            output_data[i] = np.nansum(data[prev_index:cell_index])
        elif fn == "max":
            output_data[i] = np.max(data[prev_index:cell_index])
        elif fn == "nanmax":
            output_data[i] = np.nanmax(data[prev_index:cell_index])
        elif fn == "min":
            output_data[i] = np.min(data[prev_index:cell_index])
        elif fn == "nanmin":
            output_data[i] = np.nanmin(data[prev_index:cell_index])
        else:
            raise NotImplementedError
        prev_index = cell_index
    return output_data  # ty:ignore[invalid-return-type]


@njit(cache=True)
def to_HRU(
    data: T_ArrayNumber,
    grid_to_HRU: ArrayInt32,
    land_use_ratio: ArrayFloat32,
    output_data: T_ArrayNumber,
    fn: str | None = None,
) -> T_ArrayNumber:
    """Numba helper function to convert from grid to HRU.

    Args:
        data: The grid data to be converted.
        grid_to_HRU: Array of size of the compressed grid cells. Each value maps to the index of the first unit of the next cell.
        land_use_ratio: Relative size of HRU to grid.
        output_data: Array to store the output data. Must be of size of the HRUs.
        fn: Name of function to apply to data. None if data should be directly inserted into HRUs - generally used when units are irrespective of area. 'mean' if data should first be corrected relative to the land use ratios - generally used when units are relative to area.

    Returns:
        ouput_data: Data converted to HRUs.
    """
    assert grid_to_HRU[0] != 0
    assert grid_to_HRU[-1] == land_use_ratio.size
    assert data.shape == grid_to_HRU.shape
    prev_index = 0

    if fn is None:
        for i in range(grid_to_HRU.size):
            cell_index = grid_to_HRU[i]
            output_data[prev_index:cell_index] = data[i]
            prev_index = cell_index
    elif fn == "weightedsplit":
        for i in range(grid_to_HRU.size):
            cell_index = grid_to_HRU[i]
            cell_sizes = land_use_ratio[prev_index:cell_index]
            output_data[prev_index:cell_index] = data[i] / cell_sizes.sum() * cell_sizes
            prev_index = cell_index
    else:
        raise NotImplementedError

    return output_data


class BaseVariables:
    """This class has some basic functions that can be used for variables regardless of scale."""

    mask: TwoDArrayBool

    def __init__(self) -> None:
        """Initialize BaseVariables class."""
        pass

    @property
    def shape(self) -> tuple[int, int]:
        """Get the shape of the uncompressed variable by returning the shape of the mask.

        Returns:
            shape: Shape of the uncompressed variable.
        """
        return self.mask.shape


class GridVariables(Bucket):
    """This class contains functions to handle variables on the grid scale."""

    heads: TwoDArrayFloat64
    capillar: ArrayFloat32
    layer_boundary_elevation: TwoDArrayFloat32
    elevation: ArrayFloat32
    specific_yield: TwoDArrayFloat32
    groundwater_hydraulic_conductivity_m_per_day: TwoDArrayFloat32
    upstream_area: ArrayFloat32
    upstream_area_n_cells: ArrayInt32
    river_mannings: ArrayFloat32
    river_length: ArrayFloat32
    average_river_width: ArrayFloat32
    river_alpha: ArrayFloat32
    cell_area: ArrayFloat32
    waterBodyID: ArrayInt32
    discharge_m3_s: ArrayFloat32
    discharge_in_rivers_m3_s_substep: ArrayFloat32
    waterbody_outflow_points: ArrayInt32
    interception_capacity_grassland: ThreeDArrayFloat32
    forest_crop_factor_per_10_days: ThreeDArrayFloat32
    discharge_m3_s_per_substep: TwoDArrayFloat32
    discharge_m3_s_substep: ArrayFloat32
    river_width_alpha: ArrayFloat32
    river_width_beta: ArrayFloat32
    overland_flow_buffer: TwoDArrayFloat32
    overland_flow_buffer_weights: ArrayFloat32


class Grid(BaseVariables):
    """This class is to store data in the 'normal' grid cells. This class works with compressed and uncompressed arrays. On initialization of the class, the mask of the study area is read from disk. This is the shape of any uncompressed array. Many values in this array, however, fall outside the stuy area as they are masked. Therefore, the array can be compressed by saving only the non-masked values.

    On initialization, as well as geotransformation and cell size are set, and the cell area is read from disk.

    Then, the mask is compressed by removing all masked cells, resulting in a compressed array.
    """

    var: GridVariables
    transform: Affine
    cell_area_uncompressed: TwoDArrayFloat32

    def __init__(self, data: Data, model: GEBModel) -> None:
        """Initialize Grid class.

        Args:
            data: Data class for model. Contains the various types of grids used in the GEB Model.
            model: The GEB model.
        """
        self.data = data
        self.model = model
        self.var = cast(
            GridVariables, self.model.store.create_bucket("hydrology.grid.var")
        )

        self.scaling = 1
        mask, self.transform, self.crs = read_grid(
            self.model.files["grid"]["mask"],
            return_transform_and_crs=True,
        )
        self.mask = mask.astype(bool)
        self.gt = self.transform.to_gdal()
        self.bounds = (
            self.transform.c,
            self.transform.f + self.transform.e * mask.shape[0],
            self.transform.c + self.transform.a * mask.shape[1],
            self.transform.f,
        )
        self.lon = np.linspace(
            self.transform.c + self.transform.a / 2,
            self.transform.c + self.transform.a * mask.shape[1] - self.transform.a / 2,
            mask.shape[1],
        )
        self.lat = np.linspace(
            self.transform.f + self.transform.e / 2,
            self.transform.f + self.transform.e * mask.shape[0] - self.transform.e / 2,
            mask.shape[0],
        )

        assert math.isclose(self.transform.a, -self.transform.e)
        self.cell_size = self.transform.a

        self.cell_area_uncompressed = cast(
            TwoDArrayFloat32, read_grid(self.model.files["grid"]["cell_area"])
        )

        self.mask_flat = self.mask.ravel()
        self.compressed_size = self.mask_flat.size - self.mask_flat.sum()
        self.var.cell_area = self.compress(self.cell_area_uncompressed)
        self.linear_mapping: TwoDArrayInt32 = self.create_linear_mapping(self.mask)

        BaseVariables.__init__(self)

    @property
    def lonlat(self) -> TwoDArrayFloat32:
        """Get longitude and latitude for each grid cell.

        Returns:
            lonlat: 2D array of shape (n_cells, 2) with longitude and latitude.
        """
        lonlat: TwoDArrayFloat32 = np.empty((self.compressed_size, 2), dtype=np.float32)
        lon_grid, lat_grid = np.meshgrid(self.lon, self.lat)
        lonlat[:, 0] = self.compress(lon_grid)
        lonlat[:, 1] = self.compress(lat_grid)
        return lonlat

    def create_linear_mapping(self, mask: TwoDArrayBool) -> TwoDArrayInt32:
        """Create a linear mapping from uncompressed to compressed indices.

        Returns:
            mapping: Linear mapping array.
        """
        mapping = np.full(mask.shape, -1, dtype=np.int32)
        mapping[~mask] = np.arange(self.compressed_size, dtype=np.int32)
        return mapping

    def full(self, *args: Any, **kwargs: Any) -> TwoDArrayFloat32:
        """Return a full array with size of mask. Takes any other argument normally used in np.full.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            array: Full array of mask size.
        """
        return np.full(self.mask.shape, *args, **kwargs)

    @overload
    def full_compressed(
        self, fill_value: Any, dtype: type[np.float64], **kwargs: Any
    ) -> ArrayFloat64: ...

    @overload
    def full_compressed(
        self, fill_value: Any, dtype: type[np.int32], **kwargs: Any
    ) -> ArrayInt32: ...

    @overload
    def full_compressed(
        self, fill_value: Any, dtype: type[np.int64] = ..., **kwargs: Any
    ) -> ArrayInt64: ...

    @overload
    def full_compressed(
        self, fill_value: Any, dtype: type[np.bool_], **kwargs: Any
    ) -> ArrayBool: ...

    @overload
    def full_compressed(
        self, fill_value: Any, dtype: type[np.float32] = ..., **kwargs: Any
    ) -> ArrayFloat32: ...

    def full_compressed(
        self, fill_value: Any, dtype: DTypeLike = np.float32, **kwargs: Any
    ) -> Array:
        """Return a full array with size of compressed array. Takes any other argument normally used in np.full.

        Args:
            fill_value: Value to fill the array with.
            dtype: Data type of the array. Defaults to np.float32.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            array: Full array of compressed size.
        """
        return np.full(self.compressed_size, fill_value, dtype=dtype, **kwargs)

    @overload
    def compress(self, array: TwoDArrayWithScalar) -> ArrayWithScalar: ...

    @overload
    def compress(self, array: ThreeDArrayWithScalar) -> TwoDArrayWithScalar: ...

    def compress(
        self,
        array: TwoDArrayWithScalar | ThreeDArrayWithScalar,
    ) -> ArrayWithScalar | TwoDArrayWithScalar:
        """Compress array.

        Args:
            array: Uncompressed array.

        Returns:
            array: Compressed array.
        """
        return compress(array, self.mask)

    def decompress(
        self, array: np.ndarray, fillvalue: int | float | None = None
    ) -> np.ndarray:
        """Decompress array.

        Args:
            array: Compressed array.
            fillvalue: Value to use for masked values. If None, uses NaN for float arrays and 0 for int arrays.

        Returns:
            array: Decompressed array.
        """
        return decompress_with_mask(array, self.mask, fillvalue=fillvalue)

    def plot(self, array: np.ndarray) -> None:
        """Plot array.

        Args:
            array: Array to plot.
        """
        import matplotlib.pyplot as plt

        plt.imshow(array)
        plt.show()

    def plot_compressed(
        self, array: np.ndarray, fillvalue: int | float | None = None
    ) -> None:
        """Plot compressed array.

        Args:
            array: Compressed array to plot.
            fillvalue: Value to use for masked values. If None, uses NaN for float arrays and 0 for int arrays.
        """
        self.plot(self.decompress(array, fillvalue=fillvalue))

    @overload
    def load(
        self, filepath: Path, compress: Literal[True] = True, layer: int = 1
    ) -> Array: ...

    @overload
    def load(
        self, filepath: Path, compress: Literal[True] = True, layer: None = None
    ) -> TwoDArray: ...

    @overload
    def load(
        self, filepath: Path, compress: Literal[False] = False, layer: int = 1
    ) -> TwoDArray: ...

    @overload
    def load(
        self, filepath: Path, compress: Literal[False] = False, layer: None = None
    ) -> ThreeDArray: ...

    def load(
        self, filepath: Path, compress: bool = True, layer: int | None = None
    ) -> Array | TwoDArray | ThreeDArray:
        """Load array from disk.

        Args:
            filepath: Filepath of map.
            compress: Whether to compress array.
            layer: Layer to load from file. If None, all layers are loaded. Defaults to None.

        Returns:
            array: Loaded array.
        """
        data = read_grid(filepath, layer=layer, return_transform_and_crs=False)
        assert isinstance(data, np.ndarray) and data.ndim in (2, 3)
        if compress:
            data = self.data.grid.compress(data)
        return data

    @property
    def pr_kg_per_m2_per_s(self) -> TwoDArrayFloat32:
        """Get precipitation rate for grid in kg/m²/s."""
        return self.compress(self.model.forcing.load("pr_kg_per_m2_per_s"))

    @property
    def ps_pascal(self) -> TwoDArrayFloat32:
        """Get surface pressure for grid in Pa."""
        return self.compress(self.model.forcing.load("ps_pascal"))

    @property
    def rlds_W_per_m2(self) -> TwoDArrayFloat32:
        """Get downward longwave radiation for grid in W/m²."""
        return self.compress(self.model.forcing.load("rlds_W_per_m2"))

    @property
    def rsds_W_per_m2(self) -> TwoDArrayFloat32:
        """Get downward shortwave radiation for grid in W/m²."""
        return self.compress(self.model.forcing.load("rsds_W_per_m2"))

    @property
    def tas_2m_K(self) -> TwoDArrayFloat32:
        """Get air temperature at 2m for grid in K."""
        return self.compress(self.model.forcing.load("tas_2m_K"))

    @property
    def dewpoint_tas_2m_K(self) -> TwoDArrayFloat32:
        """Get dewpoint temperature at 2m for grid in K."""
        return self.compress(self.model.forcing.load("dewpoint_tas_2m_K"))

    @property
    def wind_u10m_m_per_s(self) -> TwoDArrayFloat32:
        """Get u-component of wind at 10m for grid in m/s."""
        return self.compress(self.model.forcing.load("wind_u10m_m_per_s"))

    @property
    def wind_v10m_m_per_s(self) -> TwoDArrayFloat32:
        """Get v-component of wind at 10m for grid in m/s."""
        return self.compress(self.model.forcing.load("wind_v10m_m_per_s"))

    @property
    def spei_uncompressed(self) -> TwoDArrayFloat32:
        """Get uncompressed version of SPEI.

        We want to get the closest SPEI value, so if we are in the second
        half of the month, we want to get the first day of the next month.

        This is UNLESS we are at the end of the model run and the next
        SPEI value does not exist, in which case we want to keep using the
        last SPEI value available.
        """
        current_time: datetime = self.model.current_time

        # Determine the nearest first day of the month
        if current_time.day <= 15:
            spei_time: datetime = current_time.replace(day=1)
        else:
            # Move to the first day of the next month
            if current_time.month == 12:
                spei_time: datetime = current_time.replace(
                    year=current_time.year + 1, month=1, day=1
                )
            else:
                spei_time: datetime = current_time.replace(
                    month=current_time.month + 1, day=1
                )

            # Check if we ran out of SPEI data. If we did, revert to using the last month
            if (
                np.datetime64(spei_time, "ns")
                > self.model.forcing["SPEI"].reader.datetime_index[-1]
            ):
                spei_time: datetime = current_time.replace(day=1)

        spei: ThreeDArrayFloat32 = self.model.forcing.load(name="SPEI", dt=spei_time)
        assert spei.ndim == 3 and spei.shape[0] == 1
        spei: TwoDArrayFloat32 = spei[0]
        return spei

    @property
    def gev_c(self) -> xr.DataArray:
        """Get GEV (Generalized Extreme Value distribution) shape parameter of SPEI for grid."""
        return read_zarr(self.model.files["other"]["climate/gev_c"])

    @property
    def gev_loc(self) -> xr.DataArray:
        """Get GEV (Generalized Extreme Value distribution) location parameter of SPEI for grid."""
        return read_zarr(self.model.files["other"]["climate/gev_loc"])

    @property
    def gev_scale(self) -> xr.DataArray:
        """Get GEV (Generalized Extreme Value distribution) scale parameter of SPEI for grid."""
        return read_zarr(self.model.files["other"]["climate/gev_scale"])

    @property
    def pr_gev_c(self) -> TwoDArrayFloat32:
        """Get GEV (Generalized Extreme Value distribution) shape parameter of rainfall distribution for grid."""
        return read_grid(self.model.files["other"]["climate/pr_gev_c"])

    @property
    def pr_gev_loc(self) -> TwoDArrayFloat32:
        """Get GEV (Generalized Extreme Value distribution) location parameter of rainfall distribution for grid."""
        return read_grid(self.model.files["other"]["climate/pr_gev_loc"])

    @property
    def pr_gev_scale(self) -> TwoDArrayFloat32:
        """Get GEV (Generalized Extreme Value distribution) scale parameter of rainfall distribution for grid."""
        return read_grid(self.model.files["other"]["climate/pr_gev_scale"])


class HRUVariables(Bucket):
    """This class contains functions to handle variables on the HRU scale."""

    variable_runoff_shape_beta: ArrayFloat32
    interception_storage_m: ArrayFloat32
    snow_temperature_C: ArrayFloat32
    liquid_water_in_snow_m: ArrayFloat64
    snow_water_equivalent_m: ArrayFloat64
    topwater_m: ArrayFloat32
    reservoir_command_areas: ArrayInt32
    cell_area: ArrayFloat32
    land_use_type: ArrayInt32
    land_use_ratio: ArrayFloat32
    land_owners: ArrayInt32
    HRU_to_grid: ArrayInt32
    grid_to_HRU: ArrayInt32
    nearest_river_grid_cell: ArrayInt32
    linear_mapping: TwoDArrayInt32
    crop_age_days_map: ArrayInt32
    potential_evapotranspiration_crop_life: ArrayFloat32
    actual_evapotranspiration_crop_life: ArrayFloat32
    actual_evapotranspiration_crop_life_per_crop_stage: TwoDArrayFloat32
    potential_evapotranspiration_crop_life_per_crop_stage: TwoDArrayFloat32
    crop_map: ArrayInt32
    topwater: ArrayFloat32
    soil_layer_height_m: TwoDArrayFloat32
    water_content_field_capacity_m: TwoDArrayFloat32
    water_content_m: TwoDArrayFloat32
    water_content_saturated_m: TwoDArrayFloat32
    water_content_wilting_point_m: TwoDArrayFloat32
    water_content_residual_m: TwoDArrayFloat32
    crop_harvest_age_days: ArrayInt32
    silt_percentage: TwoDArrayFloat32
    clay_percentage: TwoDArrayFloat32
    sand_percentage: TwoDArrayFloat32
    bubbling_pressure_cm: TwoDArrayFloat32
    lambda_pore_size_distribution: TwoDArrayFloat32
    saturated_hydraulic_conductivity_m_per_s: TwoDArrayFloat32
    organic_matter_percentage: TwoDArrayFloat32
    bulk_density_kg_per_dm3: TwoDArrayFloat32
    crop_group_number_forest: ArrayFloat32
    crop_group_number_grassland_like: ArrayFloat32
    leaf_area_index_forest: TwoDArrayFloat32
    leaf_area_index_grassland_like: TwoDArrayFloat32
    interception_capacity_forest_m: TwoDArrayFloat32
    interception_capacity_grassland_like_m: TwoDArrayFloat32
    cell_length: ArrayFloat32
    water_depth_in_field: ArrayFloat32
    slope_m_per_m: ArrayFloat32
    cover: ArrayFloat32
    plant_height: ArrayFloat32
    no_erosion: ArrayBool
    tillaged: ArrayBool
    no_vegetation: ArrayBool
    stem_diameter: ArrayFloat32
    no_elements: ArrayFloat32
    canopy_cover: ArrayFloat32
    stem_diameter_harvested: ArrayFloat32
    no_elements_harvested: ArrayFloat32
    soil_enthalpy_J_per_m2: TwoDArrayFloat32
    soil_temperature_C: TwoDArrayFloat32
    solid_heat_capacity_J_per_m2_K: TwoDArrayFloat32
    solid_thermal_conductivity_W_per_m_K: TwoDArrayFloat32
    deep_soil_temperature_C: ArrayFloat32
    wetting_front_depth_m: ArrayFloat32
    wetting_front_suction_head_m: ArrayFloat32
    wetting_front_moisture_deficit: ArrayFloat32
    green_ampt_active_layer_idx: ArrayInt32
    hillslope_length_m: ArrayFloat32


class HRUs(BaseVariables):
    """This class forms the basis for the HRUs. To create the `HRUs`, each individual field owned by a farmer becomes a `HRU` first. Then, in addition, each other land use type becomes a separate HRU. `HRUs` never cross cell boundaries. This means that farmers whose fields are dispersed across multiple cells are simulated by multiple `HRUs`. Here, we assume that each `HRU`, is relatively homogeneous as it each `HRU` is operated by 1) a single farmer, or by a single other (i.e., non-farm) land-use type and 2) never crosses the boundary a hydrological model cell.

    On initalization, the mask of the study area for the cells are loaded first, and a mask on the maximum resolution of the HRUs is created. In this case, the maximum resolution of the HRUs is 20 times higher than the mask. Then the HRUs are actually created.

    Args:
        data: Data class for model.
        model: The GEB model.
    """

    var: HRUVariables

    def __init__(self, data: Data, model: GEBModel) -> None:
        """Initialize HRUs class.

        Args:
            data: Data class for model. Contains the various types of grids used in the GEB Model.
            model: The GEB model.
        """
        self.data: Data = data
        self.model: GEBModel = model

        subgrid: zarr.Group = zarr.open_group(
            self.model.files["subgrid"]["mask"], mode="r"
        )
        assert isinstance(subgrid, zarr.Group)
        y = subgrid["y"]
        x = subgrid["x"]
        assert isinstance(y, zarr.Array) and isinstance(x, zarr.Array)

        submask_height: int = y.size
        submask_width: int = x.size

        self.scaling = submask_height // self.data.grid.shape[0]
        assert submask_width // self.data.grid.shape[1] == self.scaling

        self.transform = self.data.grid.transform * Affine.scale(1 / self.scaling)
        self.crs = self.data.grid.crs

        self.gt = self.transform.to_gdal()

        self.mask = self.data.grid.mask.repeat(self.scaling, axis=0).repeat(
            self.scaling, axis=1
        )
        self.cell_size = self.data.grid.cell_size / self.scaling

        # get lats and lons for subgrid
        self.lon = np.linspace(
            self.gt[0] + self.gt[1] / 2,
            self.gt[0] + self.gt[1] * submask_width - self.gt[1] / 2,
            submask_width,
        )
        self.lat = np.linspace(
            self.gt[3] + self.gt[5] / 2,
            self.gt[3] + self.gt[5] * submask_height - self.gt[5] / 2,
            submask_height,
        )
        BaseVariables.__init__(self)

        if self.model.in_spinup:
            self.spinup()

    def spinup(self) -> None:
        """Create HRUs based on land cover and use.

        Creates the HRUs by reading the land use and farm maps and analysing them per grid cell.
        Each land use type becomes a separate HRU, and each farm field becomes a separate HRU.

        In addition, several mapping arrays are created to map between HRUs and grid cells. These are
        later used in functions to convert between HRU and grid scales.
        """
        self.var: HRUVariables = cast(
            HRUVariables,
            self.model.store.create_bucket("hydrology.HRU.var"),
        )

        (
            self.var.land_use_type,
            self.var.land_use_ratio,
            self.var.land_owners,
            self.var.HRU_to_grid,
            self.var.grid_to_HRU,
            self.var.linear_mapping,
        ) = self.create_HRUs()

        upstream_area = read_grid(self.model.files["grid"]["routing/upstream_area_m2"])

        self.var.nearest_river_grid_cell = determine_nearest_river_cell(
            upstream_area,
            self.var.HRU_to_grid,
            mask=self.data.grid.mask,
            threshold_m2=25_000_000,  # 25 km² to align with MERIT-Basins defintion of a river, https://www.reachhydro.org/home/params/merit-basins
        )

    @property
    def lonlat(self) -> TwoDArrayFloat32:
        """For each HRU, get the longitude and latitude.

        Returns:
            lonlat: 2D array of shape (n_HRUs, 2) with longitude and latitude of the grid cell the HRU is located in.
        """
        grid_lonlat: TwoDArrayFloat32 = self.data.grid.lonlat
        lonlat: TwoDArrayFloat32 = grid_lonlat[self.var.HRU_to_grid]

        return lonlat

    @property
    def linear_mapping(self) -> TwoDArrayInt32:
        """Get the linear mapping from uncompressed to compressed indices.

        Returns:
            linear_mapping: Linear mapping array.
        """
        return self.var.linear_mapping

    @property
    def compressed_size(self) -> int:
        """Gets the compressed size of a full HRU array.

        Returns:
            compressed_size: Compressed size of HRU array.
        """
        return self.var.land_use_type.size

    @staticmethod
    @njit(cache=True)
    def create_HRUs_numba(
        farms: TwoDArrayInt32,
        land_use_classes: TwoDArrayInt32,
        mask: TwoDArrayBool,
        scaling: int,
        initial_size: int,
    ) -> tuple[
        ArrayInt32,
        ArrayFloat32,
        ArrayInt32,
        ArrayInt32,
        ArrayInt32,
        TwoDArrayInt32,
    ]:
        """Numba helper function to create HRUs.

        Args:
            farms: Map of farms. Each unique integer is a unique farm. -1 is no farm.
            land_use_classes: CWatM land use class map [0-5].
            mask: Mask of the normal grid cells.
            scaling: Scaling between mask and maximum resolution of HRUs.
            initial_size: Initial size for HRU arrays.

        Returns:
            land_use_array: Land use of each HRU.
            land_use_ratio: Relative size of HRU to grid.
            land_use_owner: Owner of HRU.
            HRU_to_grid: Maps HRUs to index of compressed cell index.
            var_to_HRU: Array of size of the compressed grid cells. Each value maps to the index of the first unit of the next cell.
            linear_mapping: The index of the HRU to the subcell.
        """
        assert farms.size == mask.size * scaling * scaling
        assert farms.size == land_use_classes.size
        ysize, xsize = mask.shape

        n_nonmasked_cells = mask.size - mask.sum()
        grid_to_HRU = np.full(n_nonmasked_cells, -1, dtype=np.int32)
        linear_mapping = np.full(farms.shape, -1, dtype=np.int32)

        HRU_to_grid = np.full(initial_size, -1, dtype=np.int32)
        land_use_array = np.full(initial_size, -1, dtype=np.int32)
        land_use_size = np.full(initial_size, -1, dtype=np.int32)
        land_use_owner = np.full(initial_size, -1, dtype=np.int32)

        def resize_arrays(
            hru_to_grid: ArrayInt32,
            land_use_array: ArrayInt32,
            land_use_size: ArrayInt32,
            land_use_owner: ArrayInt32,
            new_size: int,
        ) -> tuple[ArrayInt32, ArrayInt32, ArrayInt32, ArrayInt32]:
            new_hru_to_grid = np.full(new_size, -1, dtype=np.int32)
            new_land_use_array = np.full(new_size, -1, dtype=np.int32)
            new_land_use_size = np.full(new_size, -1, dtype=np.int32)
            new_land_use_owner = np.full(new_size, -1, dtype=np.int32)

            new_hru_to_grid[: hru_to_grid.size] = hru_to_grid
            new_land_use_array[: land_use_array.size] = land_use_array
            new_land_use_size[: land_use_size.size] = land_use_size
            new_land_use_owner[: land_use_owner.size] = land_use_owner

            return (
                new_hru_to_grid,
                new_land_use_array,
                new_land_use_size,
                new_land_use_owner,
            )

        HRU = 0
        var_cell_count_compressed = 0
        var_cell_count_uncompressed = 0

        for y in range(0, ysize):
            for x in range(0, xsize):
                is_masked = mask[y, x]
                if not is_masked:
                    cell_farms = farms[
                        y * scaling : (y + 1) * scaling, x * scaling : (x + 1) * scaling
                    ].ravel()  # find farms in cell
                    cell_land_use_classes = land_use_classes[
                        y * scaling : (y + 1) * scaling, x * scaling : (x + 1) * scaling
                    ].ravel()  # get land use classes for cells
                    assert (
                        (cell_land_use_classes == 0)
                        | (cell_land_use_classes == 1)
                        | (cell_land_use_classes == 4)
                        | (cell_land_use_classes == 5)
                    ).all()

                    sort_idx = np.argsort(cell_farms)
                    cell_farms_sorted = cell_farms[sort_idx]
                    cell_land_use_classes_sorted = cell_land_use_classes[sort_idx]

                    prev_farm = -1  # farm is never -1
                    for i in range(cell_farms_sorted.size):
                        farm = cell_farms_sorted[i]
                        land_use = cell_land_use_classes_sorted[i]
                        if farm == -1:  # if area is not a farm
                            continue
                        if farm != prev_farm:
                            if HRU >= HRU_to_grid.size:
                                res_res = resize_arrays(
                                    HRU_to_grid,
                                    land_use_array,
                                    land_use_size,
                                    land_use_owner,
                                    int(HRU_to_grid.size * 1.5) + 1,
                                )
                                HRU_to_grid = res_res[0]
                                land_use_array = res_res[1]
                                land_use_size = res_res[2]
                                land_use_owner = res_res[3]

                            assert land_use_array[HRU] == -1
                            assert land_use == 1  # must be one because farm
                            land_use_array[HRU] = land_use
                            assert land_use_size[HRU] == -1
                            land_use_size[HRU] = 1
                            land_use_owner[HRU] = farm

                            HRU_to_grid[HRU] = var_cell_count_compressed

                            prev_farm = farm
                            HRU += 1
                        else:
                            land_use_size[HRU - 1] += 1

                        linear_mapping[
                            y * scaling + sort_idx[i] // scaling,
                            x * scaling + sort_idx[i] % scaling,
                        ] = HRU - 1

                    sort_idx = np.argsort(cell_land_use_classes)
                    cell_farms_sorted = cell_farms[sort_idx]
                    cell_land_use_classes_sorted = cell_land_use_classes[sort_idx]

                    prev_land_use = -1
                    assert prev_land_use != cell_land_use_classes[0]
                    for i in range(cell_farms_sorted.size):
                        land_use = cell_land_use_classes_sorted[i]
                        farm = cell_farms_sorted[i]
                        if farm != -1:
                            continue
                        if land_use != prev_land_use:
                            if HRU >= HRU_to_grid.size:
                                res_res = resize_arrays(
                                    HRU_to_grid,
                                    land_use_array,
                                    land_use_size,
                                    land_use_owner,
                                    int(HRU_to_grid.size * 1.5) + 1,
                                )
                                HRU_to_grid = res_res[0]
                                land_use_array = res_res[1]
                                land_use_size = res_res[2]
                                land_use_owner = res_res[3]

                            assert land_use_array[HRU] == -1
                            land_use_array[HRU] = land_use
                            assert land_use_size[HRU] == -1
                            land_use_size[HRU] = 1
                            prev_land_use = land_use

                            HRU_to_grid[HRU] = var_cell_count_compressed

                            HRU += 1
                        else:
                            land_use_size[HRU - 1] += 1

                        linear_mapping[
                            y * scaling + sort_idx[i] // scaling,
                            x * scaling + sort_idx[i] % scaling,
                        ] = HRU - 1

                    grid_to_HRU[var_cell_count_compressed] = HRU
                    var_cell_count_compressed += 1
                var_cell_count_uncompressed += 1

        land_use_size = land_use_size[:HRU]
        land_use_array = land_use_array[:HRU]
        land_use_owner = land_use_owner[:HRU]
        HRU_to_grid = HRU_to_grid[:HRU]
        assert int(land_use_size.sum()) == n_nonmasked_cells * scaling * scaling

        land_use_ratio = (land_use_size / (scaling**2)).astype(np.float32)
        return (
            land_use_array,
            land_use_ratio,
            land_use_owner,
            HRU_to_grid,
            grid_to_HRU,
            linear_mapping,
        )

    def create_HRUs(
        self,
    ) -> tuple[
        ArrayInt32,
        ArrayFloat32,
        ArrayInt32,
        ArrayInt32,
        ArrayInt32,
        TwoDArrayInt32,
    ]:
        """Function to create HRUs.

        Returns:
            land_use_array: Land use of each HRU.
            land_use_ratio: Relative size of HRU to grid.
            land_use_owner: Owner of HRU.
            HRU_to_grid: Maps HRUs to index of compressed cell index.
            grid_to_HRU: Array of size of the compressed grid cells. Each value maps to the index of the first unit of the next cell.
            linear_mapping: The index of the HRU to the subcell.
        """
        land_use_classes = read_grid(
            self.model.files["subgrid"]["landsurface/land_use_classes"]
        )
        initial_size = (
            np.unique(self.data.farms).size + (self.data.grid.mask == 0).sum()
        )

        res = self.create_HRUs_numba(
            self.data.farms,
            land_use_classes,
            self.data.grid.mask,
            self.scaling,
            initial_size,
        )
        return res

    def zeros(self, size: int, dtype: type, *args: Any, **kwargs: Any) -> Array:
        """Return an array of zeros with given size. Takes any other argument normally used in np.zeros.

        Args:
            size: Size of the array to create.
            dtype: Data type of the array.
            *args: Additional arguments for np.zeros.
            **kwargs: Additional keyword arguments for np.zeros.

        Returns:
            array: Array with size of number of HRUs.
        """
        return np.zeros(size, dtype, *args, **kwargs)

    @overload
    def full_compressed(
        self,
        fill_value: int,
        dtype: type[np.int32],
        *args: Any,
        **kwargs: Any,
    ) -> ArrayInt32: ...

    @overload
    def full_compressed(
        self,
        fill_value: float,
        dtype: type[np.float64],
        *args: Any,
        **kwargs: Any,
    ) -> ArrayFloat64: ...

    @overload
    def full_compressed(
        self,
        fill_value: float,
        dtype: type[np.float32],
        *args: Any,
        **kwargs: Any,
    ) -> ArrayFloat32: ...

    @overload
    def full_compressed(
        self,
        fill_value: np.float64,
        dtype: type[np.float64],
        *args: Any,
        **kwargs: Any,
    ) -> ArrayFloat64: ...

    @overload
    def full_compressed(
        self,
        fill_value: bool,
        dtype: type[np.bool_],
        *args: Any,
        **kwargs: Any,
    ) -> ArrayBool: ...

    @overload
    def full_compressed(
        self,
        fill_value: int | float | np.integer | np.floating | bool,
        dtype: type,
        *args: Any,
        **kwargs: Any,
    ) -> Array: ...

    def full_compressed(
        self,
        fill_value: int | float | np.integer | np.floating | bool,
        dtype: type,
        *args: Any,
        **kwargs: Any,
    ) -> Array:
        """Return a full array with size of number of HRUs. Takes any other argument normally used in np.full.

        Args:
            fill_value: Value to fill the array with.
            dtype: Data type of the array.
            *args: Additional arguments for np.full.
            **kwargs: Arbitrary keyword arguments for np.full.

        Returns:
            array: Array with size of number of HRUs.
        """
        return np.full(self.compressed_size, fill_value, dtype, *args, **kwargs)

    def decompress(self, HRU_array: np.ndarray) -> np.ndarray:
        """Decompress HRU array.

        Args:
            HRU_array: HRU_array.

        Returns:
            outarray: Decompressed HRU_array.
        """
        if np.issubdtype(HRU_array.dtype, np.integer):
            nanvalue = -1
        elif np.issubdtype(HRU_array.dtype, bool):
            nanvalue = False
        else:
            nanvalue = np.nan
        outarray = HRU_array[self.var.linear_mapping]
        outarray[self.mask] = nanvalue
        return outarray

    @staticmethod
    @njit(cache=True)
    def convert_subgrid_to_HRU_numba(
        array: npt.NDArray[np.generic],
        linear_mapping: npt.NDArray[np.int32],
        outarray: npt.NDArray[np.generic],
        nodatavalue: int | float | np.generic,
        method: str,
    ) -> npt.NDArray[np.generic]:
        """Numba helper function to compress subgrid array to HRU array.

        Args:
            array: Uncompressed array.
            linear_mapping: The index of the HRU to the subcell.
            outarray: Array to store the output data. Must be of size of the HRUs.
            nodatavalue: Value to use for nodata.
            method: Method to use for compression. "last" uses the last value found in the uncompressed array. "mean" uses the mean value found in the uncompressed array.

        Returns:
            outarray: Compressed array.

        Raises:
            ValueError: If method is not implemented.
        """
        array = array.ravel()
        linear_mapping = linear_mapping.ravel()
        if method == "last":
            if np.isnan(nodatavalue):
                for i in range(array.size):
                    value = array[i]
                    if not np.isnan(value):
                        HRU = linear_mapping[i]
                        outarray[HRU] = value
            else:
                for i in range(array.size):
                    value = array[i]
                    if value != nodatavalue:
                        HRU = linear_mapping[i]
                        outarray[HRU] = value
        elif method == "mean":
            array = array[linear_mapping != -1]
            linear_mapping = linear_mapping[linear_mapping != -1]
            outarray[:] = np.bincount(linear_mapping, weights=array) / np.bincount(
                linear_mapping
            )
        else:
            raise ValueError("Method not implemented")
        return outarray

    def convert_subgrid_to_HRU(
        self, array: AnyDArrayWithScalar, method: str = "last"
    ) -> AnyDArrayWithScalar:
        """Convert subgrid array to HRU array.

        Because HRUs describe multiple subgrid cells, the data within
        those subgrid cells needs to be compressed to a single value per HRU.

        It can be done in two ways:
        - "last": use the last value found in the uncompressed array.
        - "mean": use the mean value found in the uncompressed array.

        Args:
            array: Uncompressed array.
            method: Method to use for compression.
                "last" uses the last value found in the uncompressed array.
                "mean" uses the mean value found in the uncompressed array.

        Returns:
            Compressed array.
        """
        assert method in ("last", "mean"), "Only last and mean method are implemented"
        assert self.mask.shape == array.shape[-2:], "Array must have same shape as mask"
        if np.issubdtype(array.dtype, np.integer):
            fill_value = -1
        else:
            fill_value = np.nan

        output_data = np.full(
            (*array.shape[:-2], self.var.land_use_ratio.size),
            fill_value,
            dtype=array.dtype,
        )

        if array.ndim == 2:
            self.convert_subgrid_to_HRU_numba(
                array,
                self.var.linear_mapping,
                output_data,
                nodatavalue=fill_value,
                method=method,
            )
        elif array.ndim == 3:
            for i in range(array.shape[0]):
                self.convert_subgrid_to_HRU_numba(
                    array[i],
                    self.var.linear_mapping,
                    output_data[i],
                    nodatavalue=fill_value,
                    method=method,
                )
        else:
            raise NotImplementedError
        return output_data

    def plot(
        self, HRU_array: np.ndarray, ax: plt.Axes | None = None, show: bool = True
    ) -> None:
        """Function to plot HRU data.

        Args:
            HRU_array: Data to plot. Size must be equal to number of HRUs.
            ax: Optional matplotlib axis object. If given, data will be plotted on given axes.
            show: Boolean whether to show the plot or not.
        """
        import matplotlib.pyplot as plt

        assert HRU_array.size == self.compressed_size
        if ax is None:
            fig, ax = plt.subplots()
        ax.imshow(self.decompress(HRU_array), resample=False)
        if show:
            plt.show()

    @property
    def pr_kg_per_m2_per_s(self) -> TwoDArrayFloat32:
        """Get precipitation rate for HRUs in kg/m²/s."""
        pr_kg_per_m2_per_s: TwoDArrayFloat32 = self.data.grid.pr_kg_per_m2_per_s
        return self.data.to_HRU(data=pr_kg_per_m2_per_s, fn=None)

    @property
    def ps_pascal(self) -> TwoDArrayFloat32:
        """Get surface pressure for HRUs in Pa."""
        ps_pascal: TwoDArrayFloat32 = self.data.grid.ps_pascal
        return self.data.to_HRU(data=ps_pascal, fn=None)

    @property
    def rlds_W_per_m2(self) -> TwoDArrayFloat32:
        """Get downward longwave radiation for HRUs in W/m²."""
        rlds_W_per_m2: TwoDArrayFloat32 = self.data.grid.rlds_W_per_m2
        return self.data.to_HRU(data=rlds_W_per_m2, fn=None)

    @property
    def rsds_W_per_m2(self) -> TwoDArrayFloat32:
        """Get downward shortwave radiation for HRUs in W/m²."""
        rsds_W_per_m2: TwoDArrayFloat32 = self.data.grid.rsds_W_per_m2
        return self.data.to_HRU(data=rsds_W_per_m2, fn=None)

    @property
    def tas_2m_K(self) -> TwoDArrayFloat32:
        """Get air temperature at 2m for HRUs in K."""
        tas_2m_K: TwoDArrayFloat32 = self.data.grid.tas_2m_K
        return self.data.to_HRU(data=tas_2m_K, fn=None)

    @property
    def dewpoint_tas_2m_K(self) -> TwoDArrayFloat32:
        """Get dewpoint temperature at 2m for HRUs in K."""
        dewpoint_tas_2m_K: TwoDArrayFloat32 = self.data.grid.dewpoint_tas_2m_K
        return self.data.to_HRU(data=dewpoint_tas_2m_K, fn=None)

    @property
    def wind_u10m_m_per_s(self) -> TwoDArrayFloat32:
        """Get u-component of wind at 10m for HRUs in m/s."""
        wind_u10m_m_per_s: TwoDArrayFloat32 = self.data.grid.wind_u10m_m_per_s
        return self.data.to_HRU(data=wind_u10m_m_per_s, fn=None)

    @property
    def wind_v10m_m_per_s(self) -> TwoDArrayFloat32:
        """Get v-component of wind at 10m for HRUs in m/s."""
        wind_v10m_m_per_s: TwoDArrayFloat32 = self.data.grid.wind_v10m_m_per_s
        return self.data.to_HRU(data=wind_v10m_m_per_s, fn=None)


class Data:
    """The base data class for the GEB model. This class contains the data for the normal grid, the HRUs, and has methods to convert between the grid and HRUs."""

    def __init__(self, model: GEBModel) -> None:
        """Initialize Data class.

        Contains the data for the normal grid, the HRUs, and has methods to convert between the grid and HRUs.

        Args:
            model: The GEB model.
        """
        self.model = model

        self.farms = read_grid(self.model.files["subgrid"]["agents/farmers/farms"])

        self.grid = Grid(self, model)
        self.HRU = HRUs(self, model)

        if self.model.in_spinup:
            self.spinup()

    def spinup(self) -> None:
        """Spinup data class.

        Computes cell area for HRUs.
        """
        self.HRU.var.cell_area = self.to_HRU(
            data=self.grid.var.cell_area, fn="weightedsplit"
        )

    def to_HRU(
        self, *, data: T_OneorTwoDArray, fn: str | None = None
    ) -> T_OneorTwoDArray:
        """Function to convert from grid to HRU (Hydrologic Response Units).

        This method is designed to transform spatial grid data into a format suitable for HRUs, which are used in to represent distinct areas with homogeneous land use, soil type, and management conditions.

        Args:
            data (array-like): The grid data to be converted. If this parameter is set, `varname` must not be provided. Data should be an array where each element corresponds to grid cell values.
            fn (str or None): The name of the function to apply to the data before assigning it to HRUs. If `None`, the data is used as is. This is usually the case for variables that are independent of area, like temperature or precipitation fluxes. If 'weightedsplit', the data will be adjusted according to the ratios of land use within each HRU. This is important when dealing with variables that are area-dependent like precipitation or runoff volumes.

        Returns:
            output_data (array-like): Data converted to HRUs format. The structure and the type of the output depend on the input and the transformation function.

        Example:
            Suppose we have an instance of a class with a grid property containing temperature data under the attribute name 'temperature'. To convert this grid-based temperature data into HRU format, we would use:

            ```python
            temperature_HRU = instance.to_HRU(data=temperature, fn=None)
            ```

            This will fetch the temperature data from `instance.grid.temperature`, assigning the temperature to HRU within a grid cell. In other words, each HRU within a grid cell has the same temperature.

            Another example, where want to plant forest in all HRUs with grassland within an area specified by a boolean mask.

            ```python
            mask_HRU = instance.to_HRU(data=mask_grid, fn=None)
            mask_HRU[land_use_type == grass_land_use_type] = False  # set all non-grassland HRUs to False
            ```
        """
        assert not isinstance(data, list)
        # make data same size as grid, but with last dimension being size of HRU
        output_data = np.zeros(
            (*data.shape[:-1], self.HRU.var.land_use_ratio.size), dtype=data.dtype
        )

        if data.ndim == 1:
            to_HRU(
                data,
                self.HRU.var.grid_to_HRU,
                self.HRU.var.land_use_ratio,
                output_data=output_data,
                fn=fn,
            )
        elif data.ndim == 2:
            for i in range(data.shape[0]):
                to_HRU(
                    data[i],
                    self.HRU.var.grid_to_HRU,
                    self.HRU.var.land_use_ratio,
                    output_data=output_data[i],
                    fn=fn,
                )
        else:
            raise NotImplementedError
        return output_data

    def to_grid(self, *, HRU_data: np.ndarray, fn: str) -> np.ndarray:
        """Function to convert from HRUs to grid.

        Args:
            HRU_data: The HRU data to be converted (if set, varname cannot be set).
            fn: Name of function to apply to data. In most cases, several HRUs are combined into one grid unit, so a function must be applied. Choose from `mean`, `sum`, `nansum`, `max` and `min`.

        Returns:
            ouput_data: Data converted to grid units.

        Raises:
            NotImplementedError: If HRU_data is not 1D or 2D array
        """
        assert fn is not None
        assert not isinstance(HRU_data, list)
        assert isinstance(HRU_data, np.ndarray)
        if HRU_data.ndim == 1:
            outdata = to_grid(
                HRU_data,
                self.HRU.var.grid_to_HRU,
                self.HRU.var.land_use_ratio,
                fn,
            )
        elif HRU_data.ndim == 2:
            # Create output array with shape (first_dim, grid_size)
            output_data = np.zeros(
                (HRU_data.shape[0], self.HRU.var.grid_to_HRU.size),
                dtype=HRU_data.dtype,
            )
            for i in range(HRU_data.shape[0]):
                output_data[i] = to_grid(
                    HRU_data[i],
                    self.HRU.var.grid_to_HRU,
                    self.HRU.var.land_use_ratio,
                    fn,
                )
            outdata = output_data
        else:
            raise NotImplementedError("Only 1D and 2D arrays are supported")

        return outdata

    def split_HRU_data(
        self, array: AnyDArrayWithScalar, i: int, ratio: float | None = None
    ) -> AnyDArrayWithScalar:
        """Function to split HRU data.

        Args:
            array: HRU data to split.
            i: Index of HRU to split.
            ratio: Ratio of new HRU to old HRU. If None, the new HRU will have the same value as the old HRU.

        Example:
            To split HRU 5 into two HRUs, where the new HRU has 30% of the area of the old HRU, use:

            ```python
            new_HRU_data = data.split_HRU_data(old_HRU_data, 5, ratio=0.3)
            ```

            To split HRU 5 into two HRUs while keeping the original values, use:
            ```python
            new_HRU_data = data.split_HRU_data(old_HRU_data, 5)
            ```

        Returns:
            HRU data with new HRU added.
        """
        assert ratio is None or (ratio > 0 and ratio < 1)
        assert ratio is None or np.issubdtype(array.dtype, np.floating)
        assert np.issubdtype(array.dtype, (np.floating, np.integer))
        if array.ndim == 1:
            array = np.insert(array, i, array[i] * (ratio or 1), axis=0)
        elif array.ndim == 2:
            array = np.insert(array, i, array[:, i] * (ratio or 1), axis=1)
        else:
            raise NotImplementedError
        if ratio is not None:
            array[i + 1] = (1 - ratio) * array[i + 1]
        return array

    @property
    def mapping_grid_to_HRU_uncompressed(self) -> npt.NDArray[np.int32]:
        """Get uncompressed grid to HRU mapping.

        Returns:
            Uncompressed grid to HRU mapping.
        """
        return self.grid.decompress(self.HRU.var.grid_to_HRU, fillvalue=-1).ravel()

    def split(self, HRU_indices: ArrayInt32) -> int:
        """Function to split an HRU into two HRUs.

        Args:
            HRU_indices: Indices of the HRU to split. All indices must belong to the same HRU.

        Returns:
            New HRU index.

        """
        HRU = self.HRU.var.linear_mapping[HRU_indices]
        assert (
            HRU == HRU[0]
        ).all()  # assert all indices belong to same HRU - so only works for single grid cell at this moment
        HRU = HRU[0]
        assert HRU != -1

        all_HRU_indices = np.where(
            self.HRU.var.linear_mapping == HRU
        )  # this could probably be speed up
        assert (
            all_HRU_indices[0].size > HRU_indices[0].size
        )  # ensure that not all indices are split off
        ratio = HRU_indices[0].size / all_HRU_indices[0].size

        self.HRU.var.linear_mapping[self.HRU.var.linear_mapping > HRU] += 1
        self.HRU.var.linear_mapping[HRU_indices] += 1

        self.HRU.var.HRU_to_grid = self.split_HRU_data(self.HRU.var.HRU_to_grid, HRU)
        self.HRU.var.grid_to_HRU[self.HRU.var.HRU_to_grid[HRU] :] += 1

        self.HRU.var.land_owners = self.split_HRU_data(self.HRU.var.land_owners, HRU)
        self.model.agents.crop_farmers.update_field_indices()

        self.model.agents.crop_farmers.var.field_indices = self.split_HRU_data(
            self.model.agents.crop_farmers.var.field_indices, HRU
        )

        self.HRU.var.land_use_type = self.split_HRU_data(
            self.HRU.var.land_use_type, HRU
        )
        self.HRU.var.land_use_ratio = self.split_HRU_data(
            self.HRU.var.land_use_ratio, HRU, ratio=ratio
        )
        self.HRU.var.cell_area = self.split_HRU_data(
            self.HRU.var.cell_area, HRU, ratio=ratio
        )
        self.HRU.var.crop_map = self.split_HRU_data(self.HRU.var.crop_map, HRU)
        self.HRU.var.crop_age_days_map = self.split_HRU_data(
            self.HRU.var.crop_age_days_map, HRU
        )
        self.HRU.var.crop_harvest_age_days = self.split_HRU_data(
            self.HRU.var.crop_harvest_age_days, HRU
        )
        return HRU



"""Class to setup, run, and post-process the SFINCS hydrodynamic model."""

from __future__ import annotations

from collections import deque
from datetime import datetime, timedelta
from pathlib import Path
from typing import TYPE_CHECKING, Any

import geopandas as gpd
import networkx as nx
import numpy as np
import pandas as pd
import xarray as xr
from shapely.geometry.point import Point

from geb.geb_types import (
    ArrayFloat32,
    TwoDArrayFloat32,
    TwoDArrayInt32,
)
from geb.hazards.floods.workflows.utils import get_start_point
from geb.module import Module
from geb.store import Bucket
from geb.workflows.io import read_geom, read_table

from ...hydrology.landcovers import OPEN_WATER as OPEN_WATER, SEALED as SEALED
from ...workflows.io import (
    read_params,
    read_zarr,
    write_zarr,
)
from ...workflows.raster import reclassify
from .sfincs import (
    MultipleSFINCSSimulations,
    SFINCSRootModel,
    SFINCSSimulation,
)

if TYPE_CHECKING:
    from geb.model import GEBModel, Hydrology


def create_river_graph(
    rivers: gpd.GeoDataFrame, subbasins: gpd.GeoDataFrame
) -> nx.DiGraph:
    """Creates a directed graph representation of the river network.

    Args:
        rivers: A GeoDataFrame containing the river network with downstream IDs.
        subbasins: A GeoDataFrame containing the subbasins.

    Returns:
        A directed graph (networkx DiGraph) representing the river network.
    """
    rivers_without_outflow_basin = rivers[~rivers["is_downstream_outflow"]]

    river_graph: nx.DiGraph = nx.DiGraph()
    rivers_in_network = set(rivers_without_outflow_basin.index)
    for river_id, row in rivers_without_outflow_basin.iterrows():
        river_graph.add_node(river_id, uparea_m2=row["uparea_m2"])
        downstream_id = row["downstream_ID"]

        # only add edge if downstream river is in the network and not -1 (ocean)
        if downstream_id != -1 and downstream_id in rivers_in_network:
            river_graph.add_edge(river_id, downstream_id)

    return river_graph


def group_subbasins(
    river_graph: nx.DiGraph, max_area_m2: float | int
) -> dict[int, list[int]]:
    """Groups subbasins in the river graph aiming while keeping the area of each group below max_area_m2.

    Args:
        river_graph: The river network as a directed graph (networkx DiGraph).
        max_area_m2: The maximum upstream area (in m²) allowed for each group.

    Returns:
        A dictionary mapping group IDs to lists of subbasin node IDs.
    """
    river_graph = river_graph.copy()
    # add attribute for merged nodes
    nx.set_node_attributes(
        river_graph, {node: [node] for node in river_graph.nodes}, "merged_nodes"
    )

    # for each node, derive the area of the node by excluding the area of the upstream nodes
    for node in river_graph.nodes:
        upstream_nodes = list(river_graph.predecessors(node))
        upstream_area = sum(
            river_graph.nodes[up_node]["uparea_m2"] for up_node in upstream_nodes
        )
        assert upstream_area >= 0
        river_graph.nodes[node]["area_m2"] = (
            river_graph.nodes[node]["uparea_m2"] - upstream_area
        )

    groups: dict[int, list[int]] = {}
    group_id: int = 0

    while len(river_graph.nodes) > 1:
        # get all nodes without predecessors (i.e., headwater nodes)
        headwater_nodes_to_merge = [
            (n, river_graph.nodes[n]["area_m2"])
            for n, d in river_graph.in_degree()
            if d == 0
        ]
        if not headwater_nodes_to_merge:
            break

        # sort headwater nodes by area descending
        headwater_nodes_to_merge.sort(key=lambda x: x[1], reverse=True)

        for potential_node_to_merge, _ in headwater_nodes_to_merge:
            downstream_node = list(river_graph.successors(potential_node_to_merge))
            assert len(downstream_node) == 1
            downstream_node = downstream_node[0]

            new_area_after_merge = (
                river_graph.nodes[downstream_node]["area_m2"]
                + river_graph.nodes[potential_node_to_merge]["area_m2"]
            )

            if new_area_after_merge > max_area_m2:
                groups[group_id] = river_graph.nodes[potential_node_to_merge][
                    "merged_nodes"
                ]
                river_graph.remove_node(potential_node_to_merge)
                group_id += 1
                continue

            river_graph.nodes[downstream_node]["merged_nodes"].extend(
                river_graph.nodes[potential_node_to_merge]["merged_nodes"]
            )
            river_graph.nodes[downstream_node]["area_m2"] = new_area_after_merge
            # remove node
            river_graph.remove_node(potential_node_to_merge)
            break

    # add remaining nodes as groups
    for node in river_graph.nodes:
        groups[group_id] = river_graph.nodes[node]["merged_nodes"]

    return groups


class FloodVariables(Bucket):
    """Class to hold variables for the Floods module."""

    discharge_per_timestep: deque[TwoDArrayFloat32]
    runoff_m_per_timestep: deque[TwoDArrayFloat32]


class Floods(Module):
    """The class that implements all methods to setup, run, and post-process hydrodynamic flood models.

    Args:
        model: The GEB model instance.
        n_timesteps: The number of timesteps to keep in memory for discharge calculations (default is 10).
    """

    var: FloodVariables

    def __init__(self, model: GEBModel, longest_flood_event_in_days: int = 10) -> None:
        """Initializes the Floods class.

        Args:
            model: The GEB model instance.
            longest_flood_event_in_days: The number of timesteps to keep in memory for discharge calculations (default is 10).
        """
        super().__init__(model)

        self.model: GEBModel = model
        self.config: dict[str, Any] = (
            self.model.config["hazards"]["floods"]
            if "floods" in self.model.config["hazards"]
            else {}
        )

        self.DEM_config: list[dict[str, Any]] = read_params(
            self.model.files["dict"]["hydrodynamics/DEM_config"]
        )

        self.HRU = model.hydrology.HRU

        if self.model.simulate_hydrology:
            self.hydrology: Hydrology = model.hydrology
            self.longest_flood_event_in_days: int = longest_flood_event_in_days

            self.var.discharge_per_timestep: deque[ArrayFloat32] = deque(
                maxlen=self.longest_flood_event_in_days
            )
            self.var.runoff_m_per_timestep: deque[ArrayFloat32] = deque(
                maxlen=self.longest_flood_event_in_days
            )

    @property
    def name(self) -> str:
        """The name of the module."""
        return "hazard_driver.floods"

    def spinup(self) -> None:
        """Spinup method for the Floods module.

        Currently, this method does nothing as flood simulations do not require spinup.
        """
        pass

    def step(self) -> None:
        """Steps the Floods module.

        Currently, this method does nothing as flood simulations are handled in the HazardDriver.
        """
        pass

    def get_utm_zone(self, region_file: Path | str) -> str:
        """Determine the UTM zone based on the centroid of the region geometry.

        Args:
            region_file: Path to the region geometry file.

        Returns:
            The EPSG code for the UTM zone of the centroid of the region.
        """
        region: gpd.GeoDataFrame = read_geom(region_file)

        # Calculate the central longitude of the dataset
        centroid: Point = region.union_all().centroid

        # Determine the UTM zone based on the longitude
        utm_zone: int = int((centroid.x + 180) // 6) + 1

        # Determine if the data is in the Northern or Southern Hemisphere
        # The EPSG code for UTM in the northern hemisphere is EPSG:326xx (xx = zone)
        # The EPSG code for UTM in the southern hemisphere is EPSG:327xx (xx = zone)
        if centroid.y > 0:
            utm_crs: str = f"EPSG:326{utm_zone}"  # Northern hemisphere
        else:
            utm_crs: str = f"EPSG:327{utm_zone}"  # Southern hemisphere
        return utm_crs

    def build(
        self,
        name: str,
        rivers: gpd.GeoDataFrame,
        subbasins: gpd.GeoDataFrame,
        coastal: bool = False,
        low_elevation_coastal_zone_mask: gpd.GeoDataFrame | None = None,
        coastal_boundary_exclude_mask: gpd.GeoDataFrame | None = None,
        initial_water_level: float | None = 0.0,
    ) -> SFINCSRootModel:
        """Builds or reads a SFINCS model without any forcing.

        Before using this model, forcing must be set.

        When the model already exists and force_overwrite is False, the existing model is read.

        Args:
            name: Name of the SFINCS model (used for the model root directory).
            subbasins: The subbasins to build the SFINCS model for. If None, the entire model subbasins is used.
            rivers: The rivers to include in the SFINCS model.
            coastal: Whether to only include coastal areas in the model.
            low_elevation_coastal_zone_mask: A GeoDataFrame defining the low elevation coastal zone to set as active cells.
            coastal_boundary_exclude_mask: GeoDataFrame defining the areas to exclude from the coastal model boundary cells.
            initial_water_level: The initial water level to initiate the model. SFINCS fills all cells below this level with water.

        Returns:
            The built or read SFINCSRootModel instance.
        """
        sfincs_model = SFINCSRootModel(self.model.simulation_root, name)
        for entry in self.DEM_config:
            entry["elevtn"] = read_zarr(
                self.model.files["other"][entry["path"]]
            ).to_dataset(name="elevtn")

        sfincs_model.build(
            subbasins=subbasins,
            DEMs=self.DEM_config,
            rivers=rivers,
            discharge=self.discharge_spinup_ds,
            river_width_alpha=self.model.hydrology.grid.decompress(
                self.model.hydrology.grid.var.river_width_alpha
            ),
            river_width_beta=self.model.hydrology.grid.decompress(
                self.model.hydrology.grid.var.river_width_beta
            ),
            mannings=self.mannings,
            grid_size_multiplier=self.config["grid_size_multiplier"],
            subgrid=self.config["subgrid"],
            depth_calculation_method=self.model.config["hydrology"]["routing"][
                "river_depth"
            ]["method"],
            depth_calculation_parameters=self.model.config["hydrology"]["routing"][
                "river_depth"
            ]["parameters"]
            if "parameters" in self.model.config["hydrology"]["routing"]["river_depth"]
            else {},
            low_elevation_coastal_zone_mask=low_elevation_coastal_zone_mask,
            coastal_boundary_exclude_mask=coastal_boundary_exclude_mask,
            coastal=coastal,
            setup_river_outflow_boundary=not coastal,
            initial_water_level=initial_water_level,
            custom_rivers_to_burn=read_geom(
                self.model.files["geom"]["routing/custom_rivers"]
            )
            if "routing/custom_rivers" in self.model.files["geom"]
            else None,
            overwrite=self.config["overwrite"],
            p_value_threshold=self.config["p_value_threshold"],
            selection_strategy=self.config["selection_strategy"],
            fixed_shape=self.config["fixed_shape"],
            write_figures=self.config["write_figures"],
        )

        return sfincs_model

    def set_forcing(
        self,
        sfincs_model: SFINCSRootModel,
        start_time: datetime,
        end_time: datetime,
    ) -> SFINCSSimulation:
        """Sets the forcing for a SFINCS simulation.

        Depending on the forcing method in the config, either headwater point discharge
        or precipitation is used as forcing.

        Args:
            sfincs_model: The SFINCSRootModel instance to create the simulation from.
            start_time: The start time of the flood event.
            end_time: The end time of the flood event.

        Returns:
            The created SFINCSSimulation instance with the forcing set.

        Raises:
            ValueError: If the forcing method is unknown.
        """
        # Save the flood depth to a zarr file
        sfincs_simulation_name: str = f"{start_time.strftime(format='%Y%m%dT%H%M%S')} - {end_time.strftime(format='%Y%m%dT%H%M%S')}"
        if self.model.multiverse_name:
            sfincs_simulation_name: str = (
                self.model.multiverse_name + "/" + sfincs_simulation_name
            )

        simulation: SFINCSSimulation = sfincs_model.create_simulation(
            simulation_name=sfincs_simulation_name,
            start_time=start_time,
            end_time=end_time,
            write_figures=self.config["write_figures"],
            flood_map_output_interval_seconds=self.config[
                "flood_map_output_interval_seconds"
            ],
        )

        routing_substeps: int = self.var.discharge_per_timestep[0].shape[0]
        if self.config["forcing_method"] == "headwater_points":
            forcing_grid = self.hydrology.grid.decompress(
                np.vstack(
                    list(self.var.discharge_per_timestep)
                    + [
                        self.var.discharge_per_timestep[-1][-1]
                    ]  # add last timestep again (ensuring stable forcing during last hour)
                )
            )
        elif self.config["forcing_method"] == "accumulated_runoff":
            forcing_grid = self.hydrology.grid.decompress(
                np.vstack(
                    list(self.var.runoff_m_per_timestep)
                    + [self.var.runoff_m_per_timestep[-1][-1]]
                )  # add last timestep again (ensuring stable forcing during last hour)
            )
        else:
            raise ValueError(
                f"Unknown forcing method {self.config['forcing_method']}. Supported are 'headwater_points' and 'accumulated_runoff'."
            )

        substep_size: timedelta = self.model.timestep_length / routing_substeps

        # convert the forcing grid to an xarray DataArray
        forcing_grid: xr.DataArray = xr.DataArray(
            data=forcing_grid,
            coords={
                "time": pd.date_range(
                    end=self.model.current_time
                    + self.model.timestep_length,  # end of the current timestep
                    periods=len(self.var.discharge_per_timestep) * routing_substeps + 1,
                    freq=substep_size,
                ),
                "y": self.hydrology.grid.lat,
                "x": self.hydrology.grid.lon,
            },
            dims=["time", "y", "x"],
            name="forcing",
        )

        # ensure that we have forcing data for the entire event period
        assert pd.to_datetime(forcing_grid.time.values[-1]).to_pydatetime() >= end_time
        assert pd.to_datetime(forcing_grid.time.values[0]).to_pydatetime() <= start_time

        forcing_grid: xr.DataArray = forcing_grid.rio.write_crs(self.model.crs)

        if self.config["forcing_method"] == "headwater_points":
            simulation.set_headwater_forcing_from_grid(
                discharge_grid=forcing_grid,
            )

        elif self.config["forcing_method"] == "accumulated_runoff":
            river_ids: TwoDArrayInt32 = self.hydrology.grid.load(
                self.model.files["grid"]["routing/river_ids"], compress=False
            )
            basin_ids: TwoDArrayInt32 = self.hydrology.grid.load(
                self.model.files["grid"]["routing/basin_ids"], compress=False
            )
            simulation.set_accumulated_runoff_forcing(
                runoff_m=forcing_grid,
                river_network=self.model.hydrology.routing.river_network,
                river_ids=river_ids,
                basin_ids=basin_ids,
                upstream_area=self.model.hydrology.grid.decompress(
                    self.model.hydrology.grid.var.upstream_area
                ),
                cell_area=self.model.hydrology.grid.decompress(
                    self.model.hydrology.grid.var.cell_area
                ),
            )
        else:
            raise ValueError(
                f"Unknown forcing method {self.config['forcing_method']}. Supported are 'headwater_points' and 'accumulated_runoff'."
            )

        return simulation

    def run_single_event(
        self,
        start_time: datetime,
        end_time: datetime,
    ) -> None:
        """Runs a single flood event using the SFINCS model.

        Also updates the flood status of households in the model based on the flood depth results.

        Args:
            start_time: The start time of the flood event.
            end_time: The end time of the flood event.
        """
        subbasins = read_geom(self.model.files["geom"]["routing/subbasins"])
        rivers = self.model.hydrology.routing.rivers
        simulation_rivers = rivers[~rivers["is_further_downstream_outflow"]]

        river_graph = create_river_graph(simulation_rivers, subbasins)
        grouped_subbasins = group_subbasins(
            river_graph=river_graph,
            max_area_m2=1e20,  # very large to force single group only
        )

        assert len(grouped_subbasins) == 1, "currently only single group supported"
        for group_id, group in grouped_subbasins.items():
            group = set(group) | set(
                simulation_rivers.loc[simulation_rivers.index.isin(group)][
                    "downstream_ID"
                ]
            )
            subbasins_group = subbasins[subbasins.index.isin(group)]

            sfincs_root_model = self.build(
                f"group_{group_id}",
                rivers=rivers,
                subbasins=subbasins_group,
            )  # build or read the model
            sfincs_simulation = self.set_forcing(  # set the forcing
                sfincs_root_model, start_time, end_time
            )
            self.model.logger.info(
                f"Running SFINCS for {self.model.current_time}..."
            )  # log the start of the simulation

            sfincs_simulation.run(
                ncpus=self.config.get("SFINCS", {}).get("ncpus", "auto"),
                gpu=self.config.get("SFINCS", {}).get("gpu", "auto"),
            )  # run the simulation

        flood_depth: xr.DataArray = sfincs_simulation.read_max_flood_depth(
            self.config["minimum_flood_depth"]
        )  # read the flood depth results

        filename = (
            self.model.output_folder / "flood_maps" / (sfincs_simulation.name + ".zarr")
        )

        flood_depth: xr.DataArray = write_zarr(
            da=flood_depth,
            path=filename,
            crs=flood_depth.rio.crs,
        )  # save the flood depth to a zarr file

        # This check is done to compute damages (using ERA5) only after multiverse is finished
        if self.model.multiverse_name is None:
            if self.model.config["general"]["forecasts"]["use"]:
                print("Multiverse no longer active, now compute flood damages...")
            # Check if damage simulation is enabled before calculating damages
            if self.model.config["hazards"]["damage"]["simulate"]:
                self.model.agents.households.flood(flood_depth=flood_depth)

    def get_return_period_maps(self) -> None:
        """Generates flood maps for specified return periods using the SFINCS model.

        Raises:
            ValueError: If no hydrograph is found for a node and return period.
        """
        # close the zarr store
        if hasattr(self.model, "reporter"):
            self.model.reporter.variables["discharge_daily"].close()

        # load model settings
        coastal_only = self.config["coastal_only"]

        # load the subbasin geometry for the model domain
        subbasins = read_geom(self.model.files["geom"]["routing/subbasins"])
        coastal = subbasins["is_coastal"].any()

        rivers = self.model.hydrology.routing.rivers
        # if coastal load files
        if coastal:
            # Load mask of lower elevation coastal zones to activate cells for the different sfincs model regions
            low_elevation_coastal_zone_mask = read_geom(
                self.model.files["geom"]["coastal/low_elevation_coastal_zone_mask"]
            )
            low_elevation_coastal_zone_mask = low_elevation_coastal_zone_mask[
                ["initial_water_level", "geometry"]
            ]

            # use COMID as index and set unique index name for coastal region
            low_elevation_coastal_zone_mask.index = [-1]
            low_elevation_coastal_zone_mask.index.name = "COMID"

            # get initial_water_level for model domain
            initial_water_level = low_elevation_coastal_zone_mask[
                "initial_water_level"
            ].min()

            # # buffer lower elevation coastal zone mask to ensure proper inclusion of coastline
            low_elevation_coastal_zone_mask["geometry"] = (
                low_elevation_coastal_zone_mask.buffer(0.00833333)
            )

            # load osm land polygons to exclude from coastal boundary cells
            coastal_boundary_exclude_mask = read_geom(
                self.model.files["geom"]["coastal/land_polygons"],
            )

            coastal_subbasins = subbasins[subbasins["is_coastal"]].copy()

            # remove coastal subbasin from low elevation coastal zone mask
            low_elevation_coastal_zone_mask = gpd.overlay(
                low_elevation_coastal_zone_mask,
                coastal_subbasins,
                how="difference",
            )
            low_elevation_coastal_zone_mask["is_downstream_outflow"] = False
            low_elevation_coastal_zone_mask["COMID"] = 0  # 0 is not used. -1 is nan
            coastal_subbasins = pd.concat(
                [coastal_subbasins, low_elevation_coastal_zone_mask],
                ignore_index=False,
            )
            coastal_subbasins.to_file(
                "temp_coastal_subbasins.geojson", driver="GeoJSON"
            )

            model_name = "coastal_region"

            # load location and offset for coastal water level forcing
            coastal_forcing_locations: gpd.GeoDataFrame = (
                read_geom(self.model.files["geom"]["gtsm/stations_coast_rp"])
                .rename(columns={"station_id": "stations"})
                .set_index("stations")
            )  # ty:ignore[invalid-assignment]

            coastal_offset = xr.open_dataarray(
                self.model.files["other"][
                    "coastal/global_ocean_mean_dynamic_topography"
                ]
            ).rio.write_crs("EPSG:4326")

            # load sea level rise data
            sea_level_rise_rcp8p5: pd.DataFrame = read_table(
                self.model.files["table"]["gtsm/sea_level_rise_rcp8p5"]
            )

            sfincs_coastal_root_model: SFINCSRootModel = self.build(
                name=model_name,
                subbasins=coastal_subbasins,
                coastal=True,
                rivers=rivers[rivers.intersects(coastal_subbasins.union_all())],
                coastal_boundary_exclude_mask=coastal_boundary_exclude_mask,
                low_elevation_coastal_zone_mask=low_elevation_coastal_zone_mask,
                initial_water_level=initial_water_level,
            )

            if coastal_only:
                # subset subbasins to those touching the low elevation coastal zone mask
                subbasins = subbasins[
                    subbasins.intersects(low_elevation_coastal_zone_mask.union_all())
                ].copy()

        sfincs_inland_root_models: list[SFINCSRootModel] = []

        # for each subbasin, build a separate sfincs model with its downstream basin
        # when there is no downstream basin (ocean), only build for the subbasin itself
        for subbasin_id, subbasin in subbasins[
            ~subbasins["is_downstream_outflow"]
        ].iterrows():
            river: pd.Series = rivers.loc[subbasin_id]
            if not river["represented_in_grid"] and river["maxup"] == 0:
                # skip subbasins that are not represented in the grid and have no upstream area
                continue

            downstream_basin = rivers.loc[subbasin_id]["downstream_ID"]

            region_subbasins = subbasins[
                subbasins.index.isin([subbasin_id, downstream_basin])
            ].copy()
            region_rivers = rivers.copy()

            # if there is a downstream basin, mark it as downstream outflow subbasin
            if downstream_basin != -1:
                region_subbasins.at[downstream_basin, "is_downstream_outflow"] = True
                region_rivers.at[downstream_basin, "is_downstream_outflow"] = True

                sfincs_inland_root_model = self.build(
                    name=f"inland_subbasin_{subbasin_id}",
                    subbasins=region_subbasins,
                    rivers=region_rivers,
                    coastal=False,
                )
                sfincs_inland_root_model.estimate_discharge_for_return_periods(
                    discharge=self.discharge_spinup_ds,
                    return_periods=self.config["return_periods"],
                    p_value_threshold=self.config["p_value_threshold"],
                    selection_strategy=self.config["selection_strategy"],
                    fixed_shape=self.config["fixed_shape"],
                    write_figures=self.config["write_figures"],
                )
                sfincs_inland_root_models.append(sfincs_inland_root_model)

        for return_period in self.config["return_periods"]:
            simulations: list[SFINCSSimulation] = []

            if coastal:
                sfincs_coastal_simulation: SFINCSSimulation = (
                    sfincs_coastal_root_model.create_coastal_return_period_simulation(
                        return_period,
                        coastal_forcing_locations,
                        offset=coastal_offset,
                        sea_level_rise=sea_level_rise_rcp8p5,
                        year=self.model.current_time.year,
                    )
                )
                simulations.append(sfincs_coastal_simulation)

            for sfincs_inland_root_model in sfincs_inland_root_models:
                inflow_nodes = sfincs_inland_root_model.active_rivers[
                    ~sfincs_inland_root_model.active_rivers["is_downstream_outflow"]
                ]
                inflow_nodes["geometry"] = inflow_nodes["geometry"].apply(
                    get_start_point
                )

                # Build list of hydrograph DataFrames using the original node indices as column names
                Q: list[pd.DataFrame] = []
                for node_idx in inflow_nodes.index:
                    hydro = inflow_nodes.at[node_idx, f"hydrograph_{return_period}"]
                    if hydro is None:
                        raise ValueError(
                            f"No hydrograph found for node {node_idx} and return period {return_period}."
                        )
                    # hydro is expected to be dict-like {iso_timestamp: Q} — convert to DataFrame with column named node_idx
                    df = pd.DataFrame.from_dict(
                        hydro, orient="index", columns=np.array([node_idx])
                    )
                    Q.append(df)

                # Concatenate the per-node series into a single DataFrame; index -> timestamps
                Q: pd.DataFrame = pd.concat(Q, axis=1)
                Q.index = pd.to_datetime(Q.index)

                sfincs_inland_simulation: SFINCSSimulation = (
                    sfincs_inland_root_model.create_simulation(
                        simulation_name=f"rp_{return_period}",
                        start_time=Q.index[0],
                        end_time=Q.index[-1],
                    )
                )

                sfincs_inland_simulation.set_discharge_forcing_from_nodes(
                    nodes=inflow_nodes.to_crs(sfincs_inland_root_model.crs),
                    timeseries=Q,
                )

                simulations.append(sfincs_inland_simulation)

            simulation = MultipleSFINCSSimulations(simulations=simulations)

            simulation.run(
                ncpus=self.config.get("SFINCS", {}).get("ncpus", "auto"),
                gpu=self.config.get("SFINCS", {}).get("gpu", "auto"),
            )
            flood_depth_return_period: xr.DataArray = simulation.read_max_flood_depth(
                self.config["minimum_flood_depth"]
            )

            # mask floodplain with land polygons to remove inundation in the sea
            if coastal and coastal_boundary_exclude_mask is not None:
                flood_depth_return_period = flood_depth_return_period.rio.clip(
                    coastal_boundary_exclude_mask.geometry,
                    coastal_boundary_exclude_mask.crs,
                    invert=False,
                )

            write_zarr(
                flood_depth_return_period,
                self.model.output_folder / "flood_maps" / f"{return_period}.zarr",
                crs=flood_depth_return_period.rio.crs,
            )

            # simulation.cleanup()

    def run(self, event: dict[str, Any]) -> None:
        """Runs the SFINCS model for a given flood event.

        Args:
            event: A dictionary containing the flood event details, including 'start_time' and 'end_time'.
        """
        start_time = event["start_time"]
        end_time = event["end_time"]

        if self.model.config["hazards"]["floods"]["flood_risk"]:
            raise NotImplementedError(
                "Flood risk calculations are not yet implemented. Need to adapt old calculations to new flood model."
            )

        else:
            self.run_single_event(start_time, end_time)

    def save_discharge(self, discharge_m3_s_per_substep: TwoDArrayFloat32) -> None:
        """Saves the current discharge for the current timestep.

        SFINCS is run at the end of an event rather than at the beginning. Therefore,
        we need to trace back the conditions at the beginning of the event. This function
        saves the current discharge at the beginning of each timestep,
        so it can be used later when setting up the SFINCS model.
        """
        self.var.discharge_per_timestep.append(
            discharge_m3_s_per_substep
        )  # this is a deque, so it will automatically remove the oldest discharge

    def save_runoff_m(self, overland_runoff_m: TwoDArrayFloat32) -> None:
        """Saves the current runoff for the current timestep."""
        self.var.runoff_m_per_timestep.append(
            overland_runoff_m
        )  # this is a deque, so it will automatically remove the oldest runoff

    @property
    def discharge_spinup_ds(self) -> xr.DataArray:
        """Open the discharge datasets from the model output folder.

        Returns:
            The discharge data array after spinup period.

        Raises:
            ValueError: If there is not enough data available for reliable spinup.
        """
        da: xr.DataArray = read_zarr(
            self.model.output_folder
            / "report"
            / "spinup"
            / "hydrology.routing"
            / "discharge_hourly.zarr"
        )

        start_time = pd.to_datetime(da.time[0].item()) + pd.DateOffset(years=10)
        da: xr.DataArray = da.sel(time=slice(start_time, da.time[-1]))

        # make sure there is at least 20 years of data
        if len(da.time) == 0 or len(da.time.groupby(da.time.dt.year).groups) < 20:
            raise ValueError(
                """Not enough data available for reliable spinup, should be at least 20 years of data left.
                Please run the model for at least 30 years (10 years of data is discarded)."""
            )

        return da

    @property
    def mannings(self) -> xr.DataArray:
        """Get the Manning's n values for the land cover types."""
        mannings = reclassify(
            self.land_cover,
            self.land_cover_mannings_rougness_classification.set_index(
                "esa_worldcover"
            )["N"].to_dict(),
            method="lookup",
        )
        return mannings

    @property
    def land_cover(self) -> xr.DataArray:
        """Get the land cover classification for the model.

        Returns:
            An xarray DataArray containing the land cover classification.
        """
        return read_zarr(self.model.files["other"]["landcover/classification"])

    @property
    def land_cover_mannings_rougness_classification(self) -> pd.DataFrame:
        """Get the land cover classification table for Manning's roughness.

        Returns:
            A DataFrame containing the land cover classification for Manning's roughness.
        """
        return pd.DataFrame(
            data=[
                [10, "Tree cover", 10, 0.12],
                [20, "Shrubland", 20, 0.05],
                [30, "Grasland", 30, 0.034],
                [40, "Cropland", 40, 0.037],
                [50, "Built-up", 50, 0.1],
                [60, "Bare / sparse vegetation", 60, 0.023],
                [70, "Snow and Ice", 70, 0.01],
                [80, "Permanent water bodies", 80, 0.02],
                [90, "Herbaceous wetland", 90, 0.035],
                [95, "Mangroves", 95, 0.07],
                [100, "Moss and lichen", 100, 0.025],
                [0, "No data", 0, 0.1],
            ],
            columns=np.array(
                ["esa_worldcover", "description", "landuse", "N"], dtype=str
            ),
        )

    @property
    def crs(self) -> str:
        """Get the coordinate reference system (CRS) for the model.

        When the CRS is set in the configuration, it will return that value.
        If the CRS is set to "auto", it will determine the UTM zone based on the routing subbasins geometry.

        Returns:
             The CRS string, either "auto" or the determined UTM zone.
        """
        crs: str = self.config["crs"]
        if crs == "auto":
            crs: str = self.get_utm_zone(self.model.files["geom"]["routing/subbasins"])
        return crs



"""ISIMIP CO2 data adapter."""

from __future__ import annotations

from io import StringIO
from typing import Any

import pandas as pd
import requests

from .base import Adapter


class ISIMIPCO2(Adapter):
    """Adapter for ISIMIP CO2 data."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize the ISIMIP CO2 data adapter."""
        super().__init__(*args, **kwargs)

    def fetch(self, url: str) -> ISIMIPCO2:
        """Set the URL for the ISIMIP CO2 data source.

        Args:
            url: The URL of the ISIMIP CO2 data source.

        Returns:
            The current instance of the ISIMIPCO2 adapter.
        """
        self.url = url
        return self

    def read(self, scenario: str, **kwargs: Any) -> pd.DataFrame:
        """Read the ISIMIP CO2 data for a specific scenario.

        Args:
            scenario: The scenario to read (e.g., "ssp126", "ssp370", "ssp585").
            **kwargs: Additional keyword arguments (not used).

        Returns:
            A pandas DataFrame containing the CO2 concentration data with years as index and a column 'co2_ppm'.

        Raises:
            ValueError: If the scenario is invalid.
        """
        if scenario not in ["ssp126", "ssp370", "ssp585"]:
            raise ValueError(
                "Invalid scenario. Choose from 'ssp126', 'ssp370', or 'ssp585'."
            )

        historical_url: str = f"{self.url}/ISIMIP3b/InputData/climate/atmosphere_composition/co2/historical/co2_historical_annual_1850_2014.txt"
        response = requests.get(historical_url)
        response.raise_for_status()
        historical_data = response.text

        future_url: str = f"{self.url}/ISIMIP3b/InputData/climate/atmosphere_composition/co2/{scenario}/co2_{scenario}_annual_2015_2100.txt"
        response = requests.get(future_url)
        response.raise_for_status()
        future_data = response.text

        combined_data = historical_data + "\n" + future_data

        df: pd.DataFrame = pd.read_csv(
            StringIO(combined_data), sep="\s+", names=["year", "co2_ppm"]
        ).set_index("year")
        return df



"""This module contains algebraic solvers and utilities for the GEB model."""

import numpy as np
from numba import njit

from geb.geb_types import ArrayFloat


@njit(cache=True, inline="always")
def tdma_solver(
    lower_diagonal_a: ArrayFloat,
    main_diagonal_b: ArrayFloat,
    upper_diagonal_c: ArrayFloat,
    rhs_vector_d: ArrayFloat,
    solution_vector_x: ArrayFloat,
    c_prime: ArrayFloat,
    d_prime: ArrayFloat,
) -> None:
    """Solve a tridiagonal system Ax = d in-place using the Thomas algorithm.

    Notes:
        All arrays must have the same length and dtype.

    Args:
        lower_diagonal_a: Lower diagonal (length n).
        main_diagonal_b: Main diagonal (length n).
        upper_diagonal_c: Upper diagonal (length n).
        rhs_vector_d: Right hand side (length n).
        solution_vector_x: Output solution array (length n), modified in-place.
        c_prime: Work array (length n), modified in-place.
        d_prime: Work array (length n), modified in-place.
    """
    n = len(rhs_vector_d)

    c_prime[0] = upper_diagonal_c[0] / main_diagonal_b[0]
    d_prime[0] = rhs_vector_d[0] / main_diagonal_b[0]

    for i in range(1, n):
        denominator = main_diagonal_b[i] - lower_diagonal_a[i] * c_prime[i - 1]

        # Avoid division by zero
        eps = np.float32(1e-10)
        if abs(denominator) < eps:
            denominator = eps

        if i < n - 1:
            c_prime[i] = upper_diagonal_c[i] / denominator
        d_prime[i] = (
            rhs_vector_d[i] - lower_diagonal_a[i] * d_prime[i - 1]
        ) / denominator

    solution_vector_x[n - 1] = d_prime[n - 1]
    for i in range(n - 2, -1, -1):
        solution_vector_x[i] = d_prime[i] - c_prime[i] * solution_vector_x[i + 1]
